"""
Plot functions for spatial data.
"""
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.tri as mtri
import numpy as np
import scipy.spatial as sspat
import pyfar as pf
from matplotlib import colors
from mpl_toolkits.mplot3d import Axes3D
__all__ = [Axes3D]
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from packaging import version
from scipy.stats import circmean

from .cmap import phase_twilight

from pyfar.plot._utils import _add_colorbar
from pyfar.plot.utils import context as pyfar_plot_context
from spharpy.samplings import spherical_voronoi
from spharpy.plot._utils import _prepare_plot
from pyfar.classes.coordinates import sph2cart


def set_aspect_equal_3d(ax):
    """Fix equal aspect bug for 3D plots."""

    xlim = ax.get_xlim3d()
    ylim = ax.get_ylim3d()
    zlim = ax.get_zlim3d()

    from numpy import mean
    xmean = mean(xlim)
    ymean = mean(ylim)
    zmean = mean(zlim)

    plot_radius = max([abs(lim - mean_)
                       for lims, mean_ in ((xlim, xmean),
                                           (ylim, ymean),
                                           (zlim, zmean))
                       for lim in lims])

    ax.set_xlim3d([xmean - plot_radius, xmean + plot_radius])
    ax.set_ylim3d([ymean - plot_radius, ymean + plot_radius])
    ax.set_zlim3d([zmean - plot_radius, zmean + plot_radius])


def scatter(coordinates, ax=None, style='light', **kwargs):
    """Plot the x, y, and z coordinates of the sampling grid in the 3d space.

    Parameters
    ----------
    coordinates : pyfar.Coordinates, spharpy.SamplingSphere
        The coordinates to be plotted
    ax : matplotlib.axis, None, optional
        The matplotlib axis object used for plotting. By default ``None``,
        which will create a new axis object.
    style : str
        ``light`` or ``dark`` to use the pyfar plot styles
        (see :py:func:`pyfar.plot.context`) or a plot style from
        :py:data:`matplotlib.style.available`. Pass a dictionary to set
        specific plot parameters, for example
        ``style = {'axes.facecolor':'black'}``. Pass an empty dictionary
        ``style = {}`` to use the currently active plotstyle. The default is
        ``light``.
    **kwargs : optional
        Additional keyword arguments passed to the scatter function.

    Returns
    -------
    ax : matplotlib.axis
        The axis object used for plotting.

    Examples
    --------

    .. plot::

        >>> import spharpy
        >>> coords = spharpy.samplings.gaussian(n_max=5)
        >>> spharpy.plot.scatter(coords)


    """
    if not isinstance(coordinates, pf.Coordinates):
        raise ValueError("coordinates must be a coordinates object.")

    with pyfar_plot_context(style):
        fig = plt.gcf()
        if ax is None:
            ax = plt.gca() if fig.axes else plt.axes(projection='3d')

        if '3d' not in ax.name:
            raise ValueError("The projection of the axis needs to be '3d'")

        ax.scatter(coordinates.x, coordinates.y, coordinates.z, **kwargs)
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_zlabel('Z')

        ax.set_box_aspect([
            np.ptp(coordinates.x),
            np.ptp(coordinates.y),
            np.ptp(coordinates.z)])

    return ax


def _triangulation_sphere(sampling, data):
    """Triangulation for data points sampled on a spherical surface.

    Parameters
    ----------
    sampling : pyfar.Coordinates, spharpy.SamplingSphere
        Coordinate object for which the triangulation is calculated
    data : list of arrays
        x, y, and z values of the data points in the triangulation

    Returns
    -------
    triangulation : matplotlib Triangulation

    """

    x, y, z = sph2cart(
        sampling.azimuth,
        sampling.colatitude,
        np.abs(data),
        )
    hull = sspat.ConvexHull(
        np.asarray(sph2cart(
            sampling.azimuth,
            sampling.colatitude,
            np.ones(len(sampling.colatitude)))).T)
    tri = mtri.Triangulation(x, y, triangles=hull.simplices)

    return tri, [x, y, z]


def interpolate_data_on_sphere(
        sampling,
        data,
        overlap=np.pi*0.25,
        refine=False,
        interpolator='linear'):
    """Linear interpolator for data on a spherical surface. The interpolator
    exploits that the data on the sphere is periodic with regard to the
    elevation and azimuth angle. The data is periodically extended to a
    specified overlap angle before interpolation.

    Parameters
    ----------
    sampling : pyfar.Coordinates, spharpy.SamplingSphere
        The coordinates at which the data is sampled.
    data : ndarray, double
        The sampled data points.
    overlap : float, (pi/4)
        The overlap for the periodic extension in azimuth angle, given in
        radians
    refine : bool
        Refine the mesh before interpolating. The default is ``False``.
    interpolator : 'linear', 'cubic'
        The interpolation method to be used. The default is 'linear'.

    Returns
    -------
    interp : LinearTriInterpolator, CubicTriInterpolator
        The interpolator object.

    Note
    ----
    Internally, matplotlibs LinearTriInterpolator or CubicTriInterpolator
    are used.

    """
    _, lats, lons = coordinates2latlon(sampling)

    mask = lons > np.pi - overlap
    lons = np.concatenate((lons, lons[mask] - np.pi*2))
    lats = np.concatenate((lats, lats[mask]))
    data = np.concatenate((data, data[mask]))

    mask = lons < -np.pi + overlap
    lons = np.concatenate((lons, lons[mask] + np.pi*2))
    lats = np.concatenate((lats, lats[mask]))
    data = np.concatenate((data, data[mask]))

    tri = mtri.Triangulation(lons, lats)

    if refine:
        refiner = mtri.UniformTriRefiner(tri)
        tri, data = refiner.refine_field(
            data,
            triinterpolator=mtri.LinearTriInterpolator(tri, data),
            subdiv=3)

    if interpolator == 'linear':
        interpolator = mtri.LinearTriInterpolator(tri, data)
    elif interpolator == 'cubic':
        interpolator = mtri.CubicTriInterpolator(tri, data, kind='mind_E')
    else:
        raise ValueError("Please give a valid interpolation method.")

    return interpolator


def _balloon_color_data(tri, data, itype):
    """Return the data array that is to be mapped to the colormap of the
    balloon.

    Parameters
    ----------
    tri : Triangulation
        The matplotlib triangulation for the sphere
    data : ndarray, double, complex double
        The data array
    itype : 'magnitude', 'phase', 'amplitude'
        What type of data should be extracted. Either the 'magnitude', 'phase',
        or 'amplitude' of the data array is used for the colormap.

    Returns
    -------
    color_data : ndarray, double
        The data array for the colormap.
    vmin : double
        The minimum of the color data
    vmax : double
        The maximum of the color data


    """
    if itype == 'phase':
        cdata = np.mod(np.angle(data), 2*np.pi)
        vmin = 0
        vmax = 2*np.pi
        colors = circmean(cdata[tri.triangles], axis=1)
    elif itype == 'magnitude':
        cdata = np.abs(data)
        vmin = np.min(cdata)
        vmax = np.max(cdata)
        colors = np.mean(cdata[tri.triangles], axis=1)
    elif itype == 'amplitude':
        vmin = np.min(data)
        vmax = np.max(data)
        colors = np.mean(data[tri.triangles], axis=1)
    else:
        raise ValueError("Invalid type of data mapping.")

    return colors, vmin, vmax


def pcolor_sphere(
        coordinates,
        data,
        cmap=None,
        colorbar=True,
        limits=None,
        cmap_encoding='phase',
        ax=None,
        style='light',
        **kwargs):
    """Plot data on the surface of a sphere defined by the coordinate angles
    theta and phi. The data array will be mapped onto the surface of a sphere.

    Parameters
    ----------
    coordinates : pyfar.Coordinates, spharpy.SamplingSphere
        Coordinates defining a sphere
    data : ndarray, double
        Data for each angle, must have size corresponding to the number of
        points given in coordinates.
    cmap : str, :py:class:`matplotlib.colors.Colormap`, optional
        Colormap used for the plot. If ``None`` (default), the colormap
        is automatically selected based on the value of `cmap_encoding`:
        ``'phase'`` uses :py:func:`spharpy.plot.phase_twilight`, and
        ``'magnitude'`` uses the ``'viridis'`` colormap.
    colorbar : bool, optional
        Whether to show a colorbar or not. Default is ``True``.
    limits : tuple, list, optional
        Tuple or list containing the maximum and minimum to which the colormap
        needs to be clipped. If ``None``, the limits are set to the minimum and
        maximum of the data. Default is ``None``.
    cmap_encoding : str, optional
        The information encoded in the colormap. Can be either ``'phase'``
        (in radians) or ``'magnitude'``. The default is ``'phase'``.
    ax : matplotlib.axes.Axes or list, tuple or ndarray of matplotlib.axes.Axes
        Axes to plot on.

        ``None``
            Use the current axis, or create a new axis (and figure) if there is
            none.
        ``ax``
            If a single axis is passed, this is used for plotting. If
            `colorbar` is ``True`` the space for the colorbar is taken from
            this axis. The projection must be ``'3d'``.
        ``[ax, ax]``
            If a list, tuple or array of two axes is passed, the first is used
            to plot the data and the second to plot the colorbar. In this case
            `colorbar` must be ``True`` and the projection of the second axis
            must be ``'rectilinear'``. The first axis must meet the same
            conditions as for the case when a single axis is passed.

        The default is ``None``.
    style : str
        ``light`` or ``dark`` to use the pyfar plot styles
        (see :py:func:`pyfar.plot.context`) or a plot style from
        :py:data:`matplotlib.style.available`. Pass a dictionary to set
        specific plot parameters, for example
        ``style = {'axes.facecolor':'black'}``. Pass an empty dictionary
        ``style = {}`` to use the currently active plotstyle. The default is
        ``light``.
    **kwargs : optional
        Additional arguments passed to the plot_trisurf function.

    Returns
    -------
    ax : matplotlib.axes.Axes, list[matplotlib.axes.Axes]
        If `colorbar` is ``True`` a list of two axes is returned. The first
        one is the axis on which the data is plotted, the second one is the
        axis of the colorbar. If `colorbar` is ``False``, only the axis on
        which the data is plotted is returned.
    plot : matplotlib.trisurf
        The trisurf object created by the function.
    cb : matplotlib.colorbar.Colorbar, None
        The Matplotlib colorbar object if `colorbar` is ``True`` and ``None``
        otherwise. This can be used to control the appearance of the colorbar,
        e.g., the label can be set by ``colorbar.set_label()``.

    Examples
    --------

    .. plot::

        >>> import spharpy
        >>> import numpy as np
        >>> coords = spharpy.samplings.equal_area(n_max=0, n_points=500)
        >>> data = np.sin(coords.colatitude) * np.cos(coords.azimuth)
        >>> spharpy.plot.pcolor_sphere(coords, data, cmap_encoding='phase')

    """
    _check_input_parameters(coordinates, data, cmap, colorbar, limits, ax)
    if cmap_encoding not in ['phase', 'magnitude']:
        raise ValueError(
            "cmap_encoding must be either 'phase' or 'magnitude'.")

    tri, xyz = _triangulation_sphere(coordinates, np.ones_like(data))

    with pyfar_plot_context(style):
        fig, ax = _prepare_plot(ax, '3d')

        if not isinstance(ax, (list, tuple, np.ndarray)):
            ax = [ax, None]

        if '3d' not in ax[0].name:
            raise ValueError("The projection of the axis needs to be '3d'")

        if cmap_encoding == 'phase':
            if cmap is None:
                cmap = phase_twilight()
            clabel = 'Phase (rad)'
        elif cmap_encoding == 'magnitude':
            if cmap is None:
                cmap = plt.get_cmap('viridis')
            clabel = 'Magnitude'

        cdata, vmin, vmax = _balloon_color_data(tri, data, cmap_encoding)

        if limits is not None:
            vmin, vmax = limits

        plot = ax[0].plot_trisurf(tri,
                                  xyz[2],
                                  cmap=cmap,
                                  antialiased=True,
                                  vmin=vmin,
                                  vmax=vmax,
                                  **kwargs)

        plot.set_array(cdata)

        cb = _add_colorbar(colorbar, fig, ax, plot, clabel)

        # reduce to plot-axis, colorbar-axis will be returned as cb.ax
        ax = ax[0]

        ax.set_xlabel('x[m]')
        ax.set_ylabel('y[m]')
        ax.set_zlabel('z[m]')

        ax.set_box_aspect([
            np.ptp(coordinates.x),
            np.ptp(coordinates.y),
            np.ptp(coordinates.z)])

    if colorbar:
        ax = [ax, cb.ax]

    return (ax, plot, cb)


def balloon_wireframe(
        coordinates,
        data,
        cmap=None,
        colorbar=True,
        limits=None,
        cmap_encoding='phase',
        ax=None,
        style='light',
        **kwargs):
    """Plot data on a sphere defined by the coordinate angles
    theta and phi. The magnitude information is mapped onto the radius of the
    sphere. The colormap represents either the phase or the magnitude of the
    data array.

    Parameters
    ----------
    coordinates : pyfar.Coordinates, spharpy.SamplingSphere
        Coordinates defining a sphere
    data : ndarray, double
        Data for each angle, must have size corresponding to the number of
        points given in coordinates.
    cmap : str, :py:class:`matplotlib.colors.Colormap`, optional
        Colormap used for the plot. If ``None`` (default), the colormap
        is automatically selected based on the value of `cmap_encoding`:
        ``'phase'`` uses :py:func:`spharpy.plot.phase_twilight`, and
        ``'magnitude'`` uses the ``'viridis'`` colormap.
    colorbar : bool, optional
        Whether to show a colorbar or not. Default is ``True``.
    limits : tuple, list, optional
        Tuple or list containing the maximum and minimum to which the colormap
        needs to be clipped. If ``None``, the limits are set to the minimum and
        maximum of the data. Default is ``None``.
    cmap_encoding : str, optional
        The information encoded in the colormap. Can be either ``'phase'``
        (in radians) or ``'magnitude'``. The default is ``'phase'``.
    ax : matplotlib.axes.Axes or list, tuple or ndarray of matplotlib.axes.Axes
        Axes to plot on.

        ``None``
            Use the current axis, or create a new axis (and figure) if there is
            none.
        ``ax``
            If a single axis is passed, this is used for plotting. If
            `colorbar` is ``True`` the space for the colorbar is taken from
            this axis. The projection must be ``'3d'``.
        ``[ax, ax]``
            If a list, tuple or array of two axes is passed, the first is used
            to plot the data and the second to plot the colorbar. In this case
            `colorbar` must be ``True`` and the projection of the second axis
            must be ``'rectilinear'``. The first axis must meet the same
            conditions as for the case when a single axis is passed.

        The default is ``None``.
    style : str
        ``light`` or ``dark`` to use the pyfar plot styles
        (see :py:func:`pyfar.plot.context`) or a plot style from
        :py:data:`matplotlib.style.available`. Pass a dictionary to set
        specific plot parameters, for example
        ``style = {'axes.facecolor':'black'}``. Pass an empty dictionary
        ``style = {}`` to use the currently active plotstyle. The default is
        ``light``.
    **kwargs : optional
        Additional arguments passed to the plot_trisurf function.

    Returns
    -------
    ax : matplotlib.axes.Axes, list[matplotlib.axes.Axes]
        If `colorbar` is ``True`` a list of two axes is returned. The first
        one is the axis on which the data is plotted, the second one is the
        axis of the colorbar. If `colorbar` is ``False``, only the axis on
        which the data is plotted is returned.
    plot : matplotlib.trisurf
        The trisurf object created by the function.
    cb : matplotlib.colorbar.Colorbar, None
        The Matplotlib colorbar object if `colorbar` is ``True`` and ``None``
        otherwise. This can be used to control the appearance of the colorbar,
        e.g., the label can be set by ``colorbar.set_label()``.

    Examples
    --------

    .. plot::

        >>> import spharpy
        >>> import numpy as np
        >>> coords = spharpy.samplings.equal_area(n_max=0, n_points=500)
        >>> data = np.sin(coords.azimuth) * np.cos(coords.elevation)
        >>> spharpy.plot.balloon_wireframe(coords, data, cmap_encoding='phase')

    """
    # input checks
    _check_input_parameters(coordinates, data, cmap, colorbar, limits, ax)
    if cmap_encoding not in ['phase', 'magnitude']:
        raise ValueError(
            "cmap_encoding must be either 'phase' or 'magnitude'.")

    if isinstance(cmap, str):
        cmap = plt.get_cmap(cmap)

    tri, xyz = _triangulation_sphere(coordinates, data)

    with pyfar_plot_context(style):
        fig, ax = _prepare_plot(ax, '3d')

        if not isinstance(ax, (list, tuple, np.ndarray)):
            ax = [ax, None]

        if '3d' not in ax[0].name:
            raise ValueError("The projection of the axis needs to be '3d'")

        if cmap_encoding == 'phase':
            if cmap is None:
                cmap = phase_twilight()
            clabel = 'Phase (rad)'
        elif cmap_encoding == 'magnitude':
            if cmap is None:
                cmap = plt.get_cmap('viridis')
            clabel = 'Magnitude'

        cdata, vmin, vmax = _balloon_color_data(tri, data, cmap_encoding)

        if limits is not None:
            vmin, vmax = limits

        plot = ax[0].plot_trisurf(tri,
                                  xyz[2],
                                  cmap=cmap,
                                  antialiased=True,
                                  vmin=vmin,
                                  vmax=vmax,
                                  **kwargs)

        # Set values to `None`. Otherwise theses values will always overwrite
        # what is set using plot.set_facecolors.
        plot.set_array(None)

        cnorm = plt.Normalize(vmin, vmax)

        # Get colors from cdata for plot.set_edgecolors
        cmap_colors = cmap(cnorm(cdata))

        cmappable = mpl.cm.ScalarMappable(cnorm, cmap)
        cmappable.set_array(np.linspace(vmin, vmax, cdata.size))

        plot.set_edgecolors(cmap_colors)
        plot.set_facecolors(np.ones(cmap_colors.shape)*0.9)

        cb = _add_colorbar(colorbar, fig, ax, plot, clabel)

        # reduce to plot-axis, colorbar-axis will be returned as cb.ax
        ax = ax[0]

        ax.set_xlabel('x[m]')
        ax.set_ylabel('y[m]')
        ax.set_zlabel('z[m]')

        ax.set_box_aspect([
            np.ptp(xyz[0]),
            np.ptp(xyz[1]),
            np.ptp(xyz[2])])

    if colorbar:
        ax = [ax, cb.ax]

    return (ax, plot, cb)


def balloon(
        coordinates,
        data,
        cmap=None,
        colorbar=True,
        limits=None,
        cmap_encoding='phase',
        ax=None,
        style='light',
        **kwargs):
    """Plot data on a sphere defined by the coordinate angles theta and phi.
    The magnitude information is mapped onto the radius of the sphere.
    The colormap represents either the phase or the magnitude of the
    data array.

    Parameters
    ----------
    coordinates : pyfar.Coordinates, spharpy.SamplingSphere
        Coordinates defining a sphere
    data : ndarray, double
        Data for each angle, must have size corresponding to the number of
        points given in coordinates.
    cmap : str, :py:class:`matplotlib.colors.Colormap`, optional
        Colormap used for the plot. If ``None`` (default), the colormap
        is automatically selected based on the value of `cmap_encoding`:
        ``'phase'`` uses :py:func:`spharpy.plot.phase_twilight`, and
        ``'magnitude'`` uses the ``'viridis'`` colormap.
    colorbar : bool, optional
        Whether to show a colorbar or not. Default is ``True``.
    limits : tuple, list, optional
        Tuple or list containing the maximum and minimum to which the colormap
        needs to be clipped. If ``None``, the limits are set to the minimum and
        maximum of the data. Default is ``None``.
    cmap_encoding : str, optional
        The information encoded in the colormap. Can be either ``'phase'``
        (in radians) or ``'magnitude'``. The default is ``'phase'``.
    ax : matplotlib.axes.Axes or list, tuple or ndarray of matplotlib.axes.Axes
        Axes to plot on.

        ``None``
            Use the current axis, or create a new axis (and figure) if there is
            none.
        ``ax``
            If a single axis is passed, this is used for plotting. If
            `colorbar` is ``True`` the space for the colorbar is taken from
            this axis. The projection must be ``'3d'``.
        ``[ax, ax]``
            If a list, tuple or array of two axes is passed, the first is used
            to plot the data and the second to plot the colorbar. In this case
            `colorbar` must be ``True`` and the projection of the second axis
            must be ``'rectilinear'``. The first axis must meet the same
            conditions as for the case when a single axis is passed.

        The default is ``None``.
    style : str
        ``light`` or ``dark`` to use the pyfar plot styles
        (see :py:func:`pyfar.plot.context`) or a plot style from
        :py:data:`matplotlib.style.available`. Pass a dictionary to set
        specific plot parameters, for example
        ``style = {'axes.facecolor':'black'}``. Pass an empty dictionary
        ``style = {}`` to use the currently active plotstyle. The default is
        ``light``.
    **kwargs : optional
        Additional arguments passed to the plot_trisurf function.

    Returns
    -------
    ax : matplotlib.axes.Axes, list[matplotlib.axes.Axes]
        If `colorbar` is ``True`` a list of two axes is returned. The first
        one is the axis on which the data is plotted, the second one is the
        axis of the colorbar. If `colorbar` is ``False``, only the axis on
        which the data is plotted is returned.
    plot : matplotlib.trisurf
        The trisurf object created by the function.
    cb : matplotlib.colorbar.Colorbar, None
        The Matplotlib colorbar object if `colorbar` is ``True`` and ``None``
        otherwise. This can be used to control the appearance of the colorbar,
        e.g., the label can be set by ``colorbar.set_label()``.

    Examples
    --------

    .. plot::

        >>> import spharpy
        >>> import numpy as np
        >>> coords = spharpy.samplings.equal_area(n_max=0, n_points=500)
        >>> data = np.sin(coords.azimuth) * np.cos(coords.elevation)
        >>> spharpy.plot.balloon(coords, data, cmap_encoding='phase')

    """
    # input checks
    _check_input_parameters(coordinates, data, cmap, colorbar, limits, ax)
    if cmap_encoding not in ['phase', 'magnitude']:
        raise ValueError(
            "cmap_encoding must be either 'phase' or 'magnitude'.")

    tri, xyz = _triangulation_sphere(coordinates, data)

    with pyfar_plot_context(style):
        fig, ax = _prepare_plot(ax, '3d')

        # _add_colorbar expects a list of axes
        if not isinstance(ax, (list, tuple, np.ndarray)):
            ax = [ax, None]

        if '3d' not in ax[0].name:
            raise ValueError("The projection of the axis needs to be '3d'")

        if cmap_encoding == 'phase':
            if cmap is None:
                cmap = phase_twilight()
            clabel = 'Phase (rad)'
        elif cmap_encoding == 'magnitude':
            if cmap is None:
                cmap = plt.get_cmap('viridis')
            clabel = cmap_encoding.title()

        cdata, vmin, vmax = _balloon_color_data(tri, data, cmap_encoding)

        if limits is not None:
            vmin, vmax = limits

        plot = ax[0].plot_trisurf(tri,
                                  xyz[2],
                                  cmap=cmap,
                                  antialiased=True,
                                  vmin=vmin,
                                  vmax=vmax,
                                  **kwargs)

        plot.set_array(cdata)

        ax[0].set_box_aspect([
            np.ptp(xyz[0]),
            np.ptp(xyz[1]),
            np.ptp(xyz[2])])

        cb = _add_colorbar(colorbar, fig, ax, plot, clabel)

        # reduce to plot-axis, colorbar-axis will be returned as cb.ax
        ax = ax[0]

        ax.set_xlabel('x[m]')
        ax.set_ylabel('y[m]')
        ax.set_zlabel('z[m]')

    if colorbar:
        ax = [ax, cb.ax]

    return (ax, plot, cb)


def voronoi_cells_sphere(sampling, round_decimals=13, ax=None, style='light'):
    """Plot the Voronoi cells of a Voronoi tesselation on a sphere.

    Parameters
    ----------
    sampling : pyfar.Coordinates, spharpy.SamplingSphere
        Sampling as SamplingSphere object
    round_decimals : int
        Decimals to be rounded to for eliminating duplicate points in
        the voronoi diagram. Default is ``13``.
    ax : AxesSubplot, None, optional
        The subplot axes to use for plotting. The used projection needs to be
        ``'3d'``.
    style : str
        ``light`` or ``dark`` to use the pyfar plot styles
        (see :py:func:`pyfar.plot.context`) or a plot style from
        :py:data:`matplotlib.style.available`. Pass a dictionary to set
        specific plot parameters, for example
        ``style = {'axes.facecolor':'black'}``. Pass an empty dictionary
        ``style = {}`` to use the currently active plotstyle. The default is
        ``light``.

    Returns
    -------
    ax : matplotlib.axis
        The axis object used for plotting.

    Examples
    --------

    .. plot::

        >>> import spharpy
        >>> coords = spharpy.samplings.gaussian(n_max=5)
        >>> spharpy.plot.voronoi_cells_sphere(coords)

    """
    if not isinstance(sampling, pf.Coordinates):
        raise ValueError("sampling must be a coordinates object.")

    sv = spherical_voronoi(sampling, round_decimals=round_decimals)
    sv.sort_vertices_of_regions()
    points = sampling.cartesian.T

    with pyfar_plot_context(style):
        fig = plt.gcf()
        if ax is None:
            ax = plt.gca() if fig.axes else plt.axes(projection='3d')

        if '3d' not in ax.name:
            raise ValueError("The projection of the axis needs to be '3d'")

        if version.parse(mpl.__version__) < version.parse('3.1.0'):
            ax.set_aspect('equal')

        # plot the unit sphere for reference (optional)
        u = np.linspace(0, 2 * np.pi, 100)
        v = np.linspace(0, np.pi, 100)
        x = np.outer(np.cos(u), np.sin(v))
        y = np.outer(np.sin(u), np.sin(v))
        z = np.outer(np.ones(np.size(u)), np.cos(v))
        ax.plot_surface(x, y, z, color='y', alpha=0.1)

        ax.scatter(points[0], points[1], points[2], c='r')

        for region in sv.regions:
            polygon = Poly3DCollection(
                [sv.vertices[region]], alpha=0.5, facecolor=None)
            polygon.set_edgecolor((0, 0, 0, 1))
            polygon.set_facecolor((1, 1, 1, 0.))

            ax.add_collection3d(polygon)

        ax.set_box_aspect([
            np.ptp(sampling.x),
            np.ptp(sampling.y),
            np.ptp(sampling.z)])

        ax.set_xlabel('x[m]')
        ax.set_ylabel('y[m]')
        ax.set_zlabel('z[m]')

    return ax


def _combined_contour(x, y, data, limits, cmap, levels, ax):
    """Combine a filled contour plot with a black line contour plot for
    better highlighting.

    Parameters
    ----------
    x : ndarray, double
        The x coordinates.
    y : ndarray, double
        The y coordinates.
    data : ndarray, double
        The data array.
    limits : tuple, list
        Tuple or list containing the maximum and minimum to which the colormap
        needs to be clipped.
    cmap : :py:class:`matplotlib.colors.Colormap`
        Colormap for the plot, see matplotlib.cm.
    levels : int or array-like
        Determines the number and positions of the contours.
        If an int n, use :py:class:`matplotlib.ticker.MaxNLocator`,
        which tries to automatically choose
        no more than n+1 contour levels between minimum and maximum
        numeric values of the plot data. If array-like, draw contour lines at
        the specified levels. The values must be in increasing order.
    ax : matplotlib.axes
        The axes object into which the contour is plotted

    Returns
    -------
    cf : matplotlib.ScalarMappable
        The scalar mappable for the contour plot

    """
    extend = 'neither'
    if limits is None:
        limits = (data.min(), data.max())
    else:
        mask_min = data < limits[0]
        data[mask_min] = limits[0]
        mask_max = data > limits[1]
        data[mask_max] = limits[1]
        if np.any(mask_max) & np.any(mask_min):
            extend = 'both'
        elif np.any(mask_max) & ~np.any(mask_min):
            extend = 'max'
        elif ~np.any(mask_max) & np.any(mask_min):
            extend = 'min'

    ax.tricontour(x, y, data, levels=levels, linewidths=0.5, colors='k',
                  vmin=limits[0], vmax=limits[1], extend=extend)
    return ax.tricontourf(
        x, y, data, levels=levels, cmap=cmap, vmin=limits[0], vmax=limits[1],
        extend=extend)


def pcolor_map(
        coordinates,
        data,
        cmap='viridis',
        colorbar=True,
        limits=None,
        projection='mollweide',
        refine=False,
        ax=None,
        style='light',
        **kwargs):
    """
    Plot the map projection of data points sampled on a spherical surface.
    The data has to be real.

    Notes
    -----
    In case limits are given, all out of bounds data will be clipped to the
    respective limit.

    Parameters
    ----------
    coordinates : pyfar.Coordinates, spharpy.SamplingSphere
        Coordinates defining a sphere
    data : numpy.ndarray, double
        Data for each angle, must have size corresponding to the number of
        points given in coordinates.
    cmap : str, :py:class:`matplotlib.colors.Colormap`, optional
        Colormap for the plot, see matplotlib.cm. Default is ``'viridis'``.
    colorbar : bool, optional
        Whether to show a colorbar or not. Default is ``True``.
    limits : tuple, list, optional
        Tuple or list containing the maximum and minimum to which the colormap
        needs to be clipped. If ``None``, the limits are set to the minimum and
        maximum of the data. Default is ``None``.
    projection : str, optional
        The projection of the map. Default is ``'mollweide'``. See
        :py:doc:`matplotlib:gallery/subplots_axes_and_figures/geo_demo`
        for more information on available projections in matplotlib.
    refine : bool, optional
        Whether to refine the triangulation before plotting.
        Default is ``False``.
    ax : matplotlib.axes.Axes or list, tuple or ndarray of matplotlib.axes.Axes
        Axes to plot on.

        ``None``
            Use the current axis, or create a new axis (and figure) if there is
            none.
        ``ax``
            If a single axis is passed, this is used for plotting. If
            `colorbar` is ``True`` the space for the colorbar is taken from
            this axis. Projection should be one of the
            :doc:`matplotlib:gallery/subplots_axes_and_figures/geo_demo`.
        ``[ax, ax]``
            If a list, tuple or array of two axes is passed, the first is used
            to plot the data and the second to plot the colorbar. In this case
            `colorbar` must be ``True`` and the projection of the second axis
            must be ``'rectilinear'``. The first axis must meet the same
            conditions as for the case when a single axis is passed.

        The default is ``None``.
    style : str
        ``light`` or ``dark`` to use the pyfar plot styles
        (see :py:func:`pyfar.plot.context`) or a plot style from
        :py:data:`matplotlib.style.available`. Pass a dictionary to set
        specific plot parameters, for example
        ``style = {'axes.facecolor':'black'}``. Pass an empty dictionary
        ``style = {}`` to use the currently active plotstyle. The default is
        ``light``.
    **kwargs : optional
        Additional arguments passed to the tripcolor function.

    Returns
    -------
    ax : matplotlib.axes.Axes, list[matplotlib.axes.Axes]
        If `colorbar` is ``True`` a list of two axes is returned. The first
        one is the axis on which the data is plotted, the second one is the
        axis of the colorbar. If `colorbar` is ``False``, only the axis on
        which the data is plotted is returned.
    cf : matplotlib.tri.TriContourSet
        The contour plot object.
    cb : matplotlib.colorbar.Colorbar, None
        The Matplotlib colorbar object if `colorbar` is ``True`` and ``None``
        otherwise. This can be used to control the appearance of the colorbar,
        e.g., the label can be set by ``colorbar.set_label()``.

    Examples
    --------

    .. plot::

        >>> import spharpy
        >>> import numpy as np
        >>> coords = spharpy.samplings.equal_area(n_max=0, n_points=500)
        >>> data = np.sin(2*coords.colatitude) * np.cos(2*coords.azimuth)
        >>> spharpy.plot.pcolor_map(coords, data)

    """
    # input checks
    _check_input_parameters(coordinates, data, cmap, colorbar, limits, ax)
    if not isinstance(refine, bool):
        raise ValueError("refine must be a boolean.")

    height, latitude, longitude = coordinates2latlon(coordinates)
    tri = mtri.Triangulation(longitude, latitude)
    if refine is not None:
        subdiv = refine if isinstance(refine, int) else 2
        refiner = mtri.UniformTriRefiner(tri)
        tri, data = refiner.refine_field(
            data,
            triinterpolator=mtri.LinearTriInterpolator(tri, data),
            subdiv=subdiv)

    with pyfar_plot_context(style):
        fig, ax = _prepare_plot(ax, projection)

        if not isinstance(ax, (list, tuple, np.ndarray)):
            ax = [ax, None]

        if ax[0].name != projection:
            raise ValueError(
                f"The projection of the axis needs to be '{projection}'"
                f", but is '{ax[0].name}'")

        ax[0].set_xlabel('Longitude [$^\\circ$]')
        ax[0].set_ylabel('Latitude [$^\\circ$]')

        extend = 'neither'
        if limits is None:
            limits = (data.min(), data.max())
        else:
            mask_min = data < limits[0]
            data[mask_min] = limits[0]
            mask_max = data > limits[1]
            data[mask_max] = limits[1]
            if np.any(mask_max) & np.any(mask_min):
                extend = 'both'
            elif np.any(mask_max) & ~np.any(mask_min):
                extend = 'max'
            elif ~np.any(mask_max) & np.any(mask_min):
                extend = 'min'

        cf = ax[0].tripcolor(
            tri, data, cmap=cmap, vmin=limits[0], vmax=limits[1], **kwargs)

        plt.grid(True)

        cb = _add_colorbar(colorbar, fig, ax, cf, 'Amplitude')

    ax = ax[0]

    if colorbar:
        cb.extend = extend
        ax = [ax, cb.ax]

    return (ax, cf, cb)


def contour_map(
        coordinates,
        data,
        cmap='viridis',
        colorbar=True,
        limits=None,
        projection='mollweide',
        levels=None,
        ax=None,
        style='light'):
    """
    Plot the map projection of data points sampled on a spherical surface.
    The data has to be real.

    Notes
    -----
    In case limits are given, all out of bounds data will be clipped to the
    respective limit.

    Parameters
    ----------
    coordinates : pyfar.Coordinates, spharpy.SamplingSphere
        Coordinates defining a sphere
    data : ndarray, double
        Data for each angle, must have size corresponding to the number of
        points given in coordinates.
    cmap : str, :py:class:`matplotlib.colors.Colormap`, optional
        Colormap for the plot, see matplotlib.cm. Default is ``'viridis'``.
    colorbar : bool, optional
        Whether to show a colorbar or not. Default is ``True``.
    limits : tuple, list, optional
        Tuple or list containing the maximum and minimum to which the colormap
        needs to be clipped. If ``None``, the limits are set to the minimum and
        maximum of the data. Default is ``None``.
    projection : str, optional
        The projection of the map. Default is ``'mollweide'``. See
        :py:doc:`matplotlib:gallery/subplots_axes_and_figures/geo_demo`
        for more information on available projections in matplotlib.
    levels : int or array-like, optional
        Determines the number and positions of the contours.
        If an int n, use :py:class:`matplotlib.ticker.MaxNLocator`,
        which tries to automatically choose
        no more than n+1 contour levels between minimum and maximum
        numeric values of the plot data. If array-like, draw contour lines at
        the specified levels. The values must be in increasing order.
        Default is ``None``, the levels are chosen automatically by
        Matplotlib.
    ax : matplotlib.axes.Axes or list, tuple or ndarray of matplotlib.axes.Axes
        Axes to plot on.

        ``None``
            Use the current axis, or create a new axis (and figure) if there is
            none.
        ``ax``
            If a single axis is passed, this is used for plotting. If
            `colorbar` is ``True`` the space for the colorbar is taken from
            this axis. Projection should be one of the
            :doc:`matplotlib:gallery/subplots_axes_and_figures/geo_demo`.
        ``[ax, ax]``
            If a list, tuple or array of two axes is passed, the first is used
            to plot the data and the second to plot the colorbar. In this case
            `colorbar` must be ``True`` and the projection of the second axis
            must be ``'rectilinear'``. The first axis must meet the same
            conditions as for the case when a single axis is passed.

        The default is ``None``.
    style : str
        ``light`` or ``dark`` to use the pyfar plot styles
        (see :py:func:`pyfar.plot.context`) or a plot style from
        :py:data:`matplotlib.style.available`. Pass a dictionary to set
        specific plot parameters, for example
        ``style = {'axes.facecolor':'black'}``. Pass an empty dictionary
        ``style = {}`` to use the currently active plotstyle. The default is
        ``light``.

    Returns
    -------
    ax : matplotlib.axes.Axes, list[matplotlib.axes.Axes]
        If `colorbar` is ``True`` a list of two axes is returned. The first
        one is the axis on which the data is plotted, the second one is the
        axis of the colorbar. If `colorbar` is ``False``, only the axis on
        which the data is plotted is returned.
    cf : matplotlib.contour.QuadContourSet
        The contour plot object.
    cb : matplotlib.colorbar.Colorbar, None
        The Matplotlib colorbar object if `colorbar` is ``True`` and ``None``
        otherwise. This can be used to control the appearance of the colorbar,
        e.g., the label can be set by ``colorbar.set_label()``.

    Examples
    --------

    .. plot::

        >>> import spharpy
        >>> import numpy as np
        >>> coords = spharpy.samplings.equal_area(n_max=0, n_points=500)
        >>> data = np.sin(2*coords.colatitude) * np.cos(2*coords.azimuth)
        >>> spharpy.plot.contour_map(coords, data)

    """
    # input checks
    _check_input_parameters(coordinates, data, cmap, colorbar, limits, ax)
    data = data.copy()

    with pyfar_plot_context(style):
        fig, ax = _prepare_plot(ax, projection)

        if not isinstance(ax, (list, tuple, np.ndarray)):
            ax = [ax, None]

        if ax[0].name != projection:
            raise ValueError(
                f"The projection of the axis needs to be '{projection}'"
                f", but is '{ax[0].name}'")

        ax[0].set_xlabel('Longitude [$^\\circ$]')
        ax[0].set_ylabel('Latitude [$^\\circ$]')

        _, latitude, longitude = coordinates2latlon(coordinates)
        cf = _combined_contour(longitude, latitude, data, limits, cmap, levels,
                               ax[0])

        if type(levels) is int:
            levels = mpl.ticker.MaxNLocator(levels)

        plt.grid(True)

        cb = _add_colorbar(colorbar, fig, ax, cf, 'Amplitude')
        if colorbar and levels is not None:
            cb.set_ticks(levels)

    ax = ax[0]

    if colorbar:
        ax = [ax, cb.ax]

    return (ax, cf, cb)


def contour(
        coordinates,
        data,
        cmap='viridis',
        colorbar=True,
        limits=None,
        levels=None,
        ax=None,
        style='light'):
    """
    Plot the map projection of data points sampled on a spherical surface.
    The data has to be real-valued.

    Notes
    -----
    In case limits are given, all out of bounds data will be clipped to the
    respective limit.

    Parameters
    ----------
    coordinates : pyfar.Coordinates, spharpy.SamplingSphere
        Coordinates defining a sphere
    data: ndarray, double
        Data for each angle, must have size corresponding to the number of
        points given in coordinates.
    cmap : str, :py:class:`matplotlib.colors.Colormap`, optional
        Colormap for the plot, see matplotlib.cm. Default is ``'viridis'``.
    colorbar : bool, optional
        Whether to show a colorbar or not. Default is ``True``.
    limits : tuple, list, optional
        Tuple or list containing the maximum and minimum to which the colormap
        needs to be clipped. If ``None``, the limits are set to the minimum and
        maximum of the data. Default is ``None``.
    levels : int or array-like, optional
        Determines the number and positions of the contours.
        If an int n, use :py:class:`matplotlib.ticker.MaxNLocator`,
        which tries to automatically choose
        no more than n+1 contour levels between minimum and maximum
        numeric values of the plot data. If array-like, draw contour lines at
        the specified levels. The values must be in increasing order.
        Default is ``None``, the levels are chosen automatically by
        Matplotlib.
    ax : matplotlib.axes.Axes or list, tuple or ndarray of matplotlib.axes.Axes
        Axes to plot on.

        ``None``
            Use the current axis, or create a new axis (and figure) if there is
            none.
        ``ax``
            If a single axis is passed, this is used for plotting. If
            `colorbar` is ``True`` the space for the colorbar is taken from
            this axis. The projection must be ``'rectilinear'``.
        ``[ax, ax]``
            If a list, tuple or array of two axes is passed, the first is used
            to plot the data and the second to plot the colorbar. In this case
            `colorbar` must be ``True`` and the projection of the second axis
            must be ``'rectilinear'``. The first axis must meet the same
            conditions as for the case when a single axis is passed.

        The default is ``None``.
    style : str
        ``light`` or ``dark`` to use the pyfar plot styles
        (see :py:func:`pyfar.plot.context`) or a plot style from
        :py:data:`matplotlib.style.available`. Pass a dictionary to set
        specific plot parameters, for example
        ``style = {'axes.facecolor':'black'}``. Pass an empty dictionary
        ``style = {}`` to use the currently active plotstyle. The default is
        ``light``.

    Returns
    -------
    ax : matplotlib.axes.Axes, list[matplotlib.axes.Axes]
        If `colorbar` is ``True`` a list of two axes is returned. The first
        one is the axis on which the data is plotted, the second one is the
        axis of the colorbar. If `colorbar` is ``False``, only the axis on
        which the data is plotted is returned.
    cf : matplotlib.contour.QuadContourSet
        The contour plot object.
    cb : matplotlib.colorbar.Colorbar, None
        The Matplotlib colorbar object if `colorbar` is ``True`` and ``None``
        otherwise. This can be used to control the appearance of the colorbar,
        e.g., the label can be set by ``colorbar.set_label()``.

    Examples
    --------

    .. plot::

        >>> import spharpy
        >>> import numpy as np
        >>> coords = spharpy.samplings.equal_area(n_max=0, n_points=500)
        >>> data = np.sin(2*coords.colatitude) * np.cos(2*coords.azimuth)
        >>> spharpy.plot.contour(coords, data)

    """
    # input checks
    _check_input_parameters(coordinates, data, cmap, colorbar, limits, ax)
    data = data.copy()

    _, latitude, longitude = coordinates2latlon(coordinates)
    lat_deg = latitude * 180/np.pi
    lon_deg = longitude * 180/np.pi

    with pyfar_plot_context(style):
        fig, ax = _prepare_plot(ax, 'rectilinear')

        if not isinstance(ax, (list, tuple, np.ndarray)):
            ax = [ax, None]

        if ax[0].name != 'rectilinear':
            raise ValueError(
                f"The projection of the axis needs to be 'rectilinear'"
                f", but is '{ax[0].name}'")

        ax[0].set_xlabel('Longitude [$^\\circ$]')
        ax[0].set_ylabel('Latitude [$^\\circ$]')

        cf = _combined_contour(lon_deg, lat_deg, data, limits, cmap, levels,
                               ax[0])

        if type(levels) is int:
            levels = mpl.ticker.MaxNLocator(levels)

        plt.grid(True)

        cb = _add_colorbar(colorbar, fig, ax, cf, 'Amplitude')
        if colorbar and levels is not None:
            cb.set_ticks(levels)

    ax = ax[0]

    if colorbar:
        ax = [ax, cb.ax]

    return (ax, cf, cb)


class MidpointNormalize(colors.Normalize):
    """Colormap norm with a defined midpoint. Useful for normalization of
    colormaps representing deviations from a defined midpoint.
    Taken from the official matplotlib documentation at
    https://matplotlib.org/users/colormapnorms.html.
    """

    def __init__(self, vmin=None, vmax=None, midpoint=0., clip=False):
        self.midpoint = midpoint
        colors.Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):  # noqa: ARG002
        # I'm ignoring masked values and all kinds of edge cases to make a
        # simple example...
        x, y = [self.vmin, self.midpoint, self.vmax], [0, 0.5, 1]
        return np.ma.masked_array(np.interp(value, x, y))


def coordinates2latlon(coords: pf.Coordinates):
    r"""Transforms from Cartesian coordinates to Geocentric coordinates.

    .. math::

        h = \sqrt{x^2 + y^2 + z^2},

        \theta = \pi/2 - \arccos(\frac{z}{r}),

        \phi = \arctan(\frac{y}{x})

        -\pi/2 < \theta < \pi/2,

        -\pi < \phi < \pi

    where :math:`h` is the heigth, :math:`\theta` is the latitude angle
    and :math:`\phi` is the longitude angle

    Parameters
    ----------
    coords : pyfar.Coordinates, spharpy.SamplingSphere
        Cartesian Coordiantes are transformed to Geocentric coordinates

    Returns
    -------
    height : ndarray, number
        The radius is rendered as height information
    latitude : ndarray, number
        Geocentric latitude angle
    longitude : ndarray, number
        Geocentric longitude angle

    """
    x = coords.x
    y = coords.y
    z = coords.z
    height = np.sqrt(x**2 + y**2 + z**2)
    latitude = np.pi/2 - np.arccos(z/height)
    longitude = np.arctan2(y, x)
    return height, latitude, longitude


def _check_input_parameters(coordinates, data, cmap, colorbar, limits,
                            ax=None):
    """Check the input parameters for the plotting functions.

    The function raises ValueError if the input parameters are not valid.

    Parameters
    ----------
    coordinates : pyfar.Coordinates, spharpy.SamplingSphere
        Coordinates defining a sphere
    data : ndarray, double
        Data for each angle, must have size corresponding to the number of
        points given in coordinates.
    cmap : str, :py:class:`matplotlib.colors.Colormap`
        Colormap for the plot, see matplotlib.cm.
    colorbar : bool, optional
        Whether to show a colorbar or not. Default is `True`.
    limits : tuple, list, optional
        Tuple or list containing the maximum and minimum to which the colormap
        needs to be clipped. If `None`, the limits are set to the minimum and
        maximum of the data.
    ax : matplotlib.axes.Axes or list, tuple or ndarray of matplotlib.axes.Axes
        Axes to plot on.
    """
    if not isinstance(colorbar, bool):
        raise ValueError("colorbar must be a boolean.")
    if not isinstance(cmap, (str, type(None), mpl.colors.Colormap)):
        raise ValueError(
            "cmap must be a string, Colormap object, or None.")
    if not isinstance(coordinates, pf.Coordinates):
        raise ValueError("coordinates must be a coordinates object.")
    if not isinstance(data, np.ndarray):
        raise ValueError(
            "data must be a 1D array with the same cshape as the coordinates.")
    if data.shape[-1] != coordinates.cshape[-1]:
        raise ValueError(
            "data must be a 1D array with the same cshape as the coordinates.")
    if limits is not None and not isinstance(limits, (tuple, list)):
        raise ValueError(
            "limits must be a tuple or list containing the minimum and "
            "maximum values for the colormap or None.")
    if limits is not None and len(limits) != 2:
        raise ValueError(
            "limits must be a tuple or list containing the minimum and "
            "maximum values for the colormap or None.")
    if not colorbar and isinstance(ax, (tuple, list, np.ndarray)):
        raise ValueError(
            "A list of axes can not be used if colorbar is False")
    if not (ax is None or isinstance(ax, (list, tuple, np.ndarray, plt.Axes)))\
        or (isinstance(ax, (list, tuple, np.ndarray))
            and np.asarray(ax).shape != (2,)):
        raise ValueError(
            "ax can be ``None``, a single matplotlib.axes.Axes object or a "
            "list, tuple or array of two axes",
        )
    if isinstance(ax, (tuple, list, np.ndarray)) \
            and (ax[1].name != 'rectilinear'):
        raise ValueError(
            "If [ax1, ax2] is passed ax2 needs to be of 'rectilinear' "
            "projection",
        )



"""spharpy classes."""

from .sh import (
    SphericalHarmonicDefinition,
    SphericalHarmonics,
)

from .audio import (
    SphericalHarmonicSignal,
    SphericalHarmonicTimeData,
    SphericalHarmonicFrequencyData,
)

from .coordinates import (
    SamplingSphere,
)

__all__ = [
    'SphericalHarmonicDefinition',
    'SphericalHarmonics',
    'SphericalHarmonicSignal',
    'SphericalHarmonicTimeData',
    'SphericalHarmonicFrequencyData',
    'SamplingSphere',
]



# -*- coding: utf-8 -*-

"""Top-level package for spharpy."""

__author__ = """The pyfar developers"""
__email__ = 'info@pyfar.org'
__version__ = '1.0.0'

from .classes.sh import (
    SphericalHarmonics,
    SphericalHarmonicDefinition,
)
from .classes.coordinates import SamplingSphere
from .classes.audio import (
    SphericalHarmonicSignal,
    SphericalHarmonicTimeData,
    SphericalHarmonicFrequencyData,
)
from . import spherical
from . import samplings
from . import plot
from . import transforms
from . import beamforming
from . import interpolate
from . import spatial
from . import special


__all__ = [
    'SphericalHarmonics',
    'SphericalHarmonicDefinition',
    'SphericalHarmonicSignal',
    'SphericalHarmonicTimeData',
    'SphericalHarmonicFrequencyData',
    'spherical',
    'samplings',
    'plot',
    'transforms',
    'beamforming',
    'interpolate',
    'spatial',
    'special',
    'SamplingSphere',
]



"""Interpolation module for spherical data."""

from scipy import interpolate as spinterpolate
import numpy as np


class SmoothSphereBivariateSpline(spinterpolate.SmoothSphereBivariateSpline):
    """Smooth bivariate spline approximation in spherical coordinates.
    The implementation uses the method proposed by Dierckx [#]_.

    Parameters
    ----------
    sampling : :py:class:`pyfar.Coordinates`
        Coordinates object containing the positions for which the data
        is sampled
    data : array, float
        Array containing the data at the sampling positions. Has to be
        real-valued.
    w : array, float
        Weighting coefficients. The default is ``None``, which sets the
        weighting coefficients to ``1``.
    s : float, 1e-4
        Smoothing factor > 0
        Positive smoothing factor defined for estimation condition:
        ``sum((w(i)*(r(i) - s(theta(i), phi(i))))**2, axis=0) <= s``
        Default ``s=len(w)`` which should be a good value if ``1/w[i]`` is
        an estimate of the standard deviation of ``r[i]``. The default
        value is ``1e-4``
    eps : float, 1e-16
        Sets the threshold used to determine the effective rank of the
        underlying linear system of equations. Values smaller than this
        threshold are treated as zero during the fitting process.
        Depends on the used data type and numerical precision. The default
        is ``1e-16``.

    Note
    ----
    This is a wrapper for scipy's SmoothSphereBivariateSpline only using
    Coordinates objects as container for the azimuth and elevation angles.
    For detailed information see scipy's documentation.

    References
    ----------
    .. [#] P. Dierckx, “Algorithms for smoothing data on the sphere with
           tensor product splines,” p. 24, 1984.

    Examples
    --------
    >>> n_max = 10
    >>> sampling = spharpy.samplings.equalarea(
    ...     n_max, n_points=500, condition_num=np.inf)
    >>> Y = spharpy.spherical.spherical_harmonic_basis_real(n_max, sampling)
    >>> y_vec = spharpy.spherical.spherical_harmonic_basis_real(
    ...     n_max, Coordinates(1, 0, 0))
    >>> data = Y @ y_vec.T
    >>> data = np.sin(sampling.elevation)*np.sin(2*sampling.azimuth)
    >>> interp_grid = spharpy.samplings.hyperinterpolation(35)
    >>> data_grid = np.sin(interp_grid.elevation)*np.sin(2*interp_grid.azimuth)
    >>> interpolator = interpolate.SmoothSphereBivariateSpline(
    ...     sampling, data, s=1e-4)
    ...
    >>> interp_data = interpolator(interp_grid)

    """ # noqa: 501

    def __init__(self, sampling, data, w=None, s=1e-4, eps=1e-16):
        theta = sampling.colatitude
        phi = sampling.azimuth
        if np.any(np.iscomplex(data)):
            raise ValueError("Complex data is not supported.")
        super().__init__(theta, phi, data, w, s, eps)

    def __call__(self, interp_grid, dtheta=0, dphi=0):
        """Evaluate the spline on a new sampling grid.

        Parameters
        ----------
        interp_grid : :py:class:`pyfar.Coordinates`
            Coordinates object containing a new set of points for which data
            is to be interpolated.
        dtheta : int, optional
            Order of theta derivative. Default is ``0``.
        dphi : int, optional
            Order of phi derivative. Default is ``0``.

        """ # noqa: 501
        theta = interp_grid.colatitude
        phi = interp_grid.azimuth
        return super().__call__(
            theta, phi, dtheta=dtheta, dphi=dphi, grid=False)

    def get_coeffs(self):
        """Get the coefficients of the spline.

        Returns
        -------
        coeffs : ndarray, float
            The coefficients of the spline. The shape is
            (n_coeffs_theta, n_coeffs_phi, n_coefficients).
            The coefficients are ordered by increasing degree and order.
        """
        return super().get_coeffs()

    def get_knots(self):
        """Get the knots of the spline.

        Returns
        -------
        knots : tuple of ndarray, float
            The knots of the spline. The first element is the knots in theta
            direction and the second element is the knots in phi direction.
        """
        return super().get_knots()

    def get_residual(self):
        """Get the residual of the spline.

        Returns
        -------
        residual : float
            The residual of the spline, which is the sum of squared errors
            between the data and the spline evaluation.
        """
        return super().get_residual()



"""
Module implementing or wrapping special functions required for higher level
classes and functions of the spharpy package.
"""

from itertools import count
import numpy as np
import scipy.special as _spspecial
from scipy.optimize import brentq


def spherical_bessel(n, z, derivative=False):
    r"""
    Spherical Bessel function of the first kind and of order n evaluated at z.

    The spherical Bessel function of the first kind is defined as
    ([#]_, Eq. 2.29)

    .. math::

        j_n(z) = \sqrt{\frac{\pi}{2z}} J_{n+\frac{1}{2}} (z)

    with the Bessel function of the first kind :math:`J_\alpha`.

    Parameters
    ----------
    n : int, ndarray
        Order of the spherical Bessel function
    z : double, ndarray
        Argument of the spherical Bessel function. Has to be real valued.
    derivative : bool
        Return the derivative of the spherical Bessel function. Default is
        ``False``.


    Returns
    -------
    jn : double, ndarray
        Spherical Bessel function. Array with dimensions [N x Z], where N is
        the number of elements in n and Z is the number of elements in z.

    Note
    ----
    This is a wrapper around the Scipy implementation of the
    :py:func:`spherical Bessel function <scipy.special.spherical_jn>`.

    References
    ----------
    .. [#] Rafaely, B. (2019). Fundamentals of spherical array processing
           (2nd ed.). Springer. https://doi.org/10.1007/978-3-319-99561-8
    """

    ufunc = _spspecial.spherical_jn
    n = np.asarray(n, dtype=int)
    z = np.asarray(z, dtype=float)

    bessel = np.zeros((n.size, z.size), dtype=complex)

    if n.size > 1:
        for idx, order in zip(count(), n):
            bessel[idx, :] = ufunc(order, z, derivative=derivative)
    else:
        bessel = ufunc(n, z, derivative=derivative)

    if z.ndim <= 1 or n.ndim <= 1:
        bessel = np.squeeze(bessel)

    return bessel


def spherical_bessel_zeros(n_max, n_zeros):
    r"""Compute the zeros of the spherical Bessel function.

    This function will always start at order zero which is equal
    to :math:`\sin(z)/z` and iteratively compute the roots for higher orders.
    The roots are computed using Brents algorithm from the scipy package. See
    :py:func:`~spherical_bessel` for more information.

    Parameters
    ----------
    n_max : int
        The order of the spherical bessel function
    n_zeros : int
        The number of roots to be computed

    Returns
    -------
    roots : ndarray, double
        The roots of the spherical bessel function

    """
    def func(x, n):
        return _spspecial.spherical_jn(n, x)

    zerosj = np.zeros((n_max+1, n_zeros), dtype=float)
    zerosj[0] = np.arange(1, n_zeros+1)*np.pi
    points = np.arange(1, n_zeros+n_max+1)*np.pi

    roots = np.zeros(n_zeros+n_max, dtype=float)
    for i in range(1, n_max+1):
        for j in range(n_zeros+n_max-i):
            roots[j] = brentq(func, points[j], points[j+1], (i,), maxiter=5000)
        points = roots
        zerosj[i, :n_zeros] = roots[:n_zeros]

    return zerosj


def spherical_hankel(n, z, kind=2, derivative=False):
    r"""
    Spherical Hankel function of the first kind and order n evaluated at z.

    The spherical Hankel function of the first kind is defined as
    ([#]_, Eq. 2.30)

    .. math::

        h_n(z) = \sqrt{\frac{\pi}{2z}} H_{n+\frac{1}{2}} (z)

    with the Hankel function of the first kind :math:`H_\alpha`.

    Parameters
    ----------
    n : int, ndarray
        Order of the spherical bessel function
    z : double, ndarray
        Argument of the spherical bessel function. Has to be real valued.
    kind : int, optional
        The kind of the spherical Hankel function. Must be ``1`` or ``2``. The
        default is ``2``.
    derivative : bool, optional
        Flag to indicate if the derivative os the spherical Hankel function
        should be returned. The default is ``False``.

    Returns
    -------
    hn : double, ndarray
        Spherical bessel function. Array with dimensions [N x Z], where N is
        the number of elements in n and Z is the number of elements in z.

    References
    ----------
    .. [#] Rafaely, B. (2019). Fundamentals of spherical array processing
           (2nd ed.). Springer. https://doi.org/10.1007/978-3-319-99561-8

    Note
    ----
    This is based on the Hankel functions implemented in the scipy package.
    """

    if kind not in (1, 2):
        raise ValueError("The spherical hankel function can \
            only be of first or second kind.")

    n = np.asarray(n, dtype=int)
    z = np.asarray(z, dtype=float)

    ufunc = _spherical_hankel_derivative if derivative else _spherical_hankel

    if n.size > 1:
        hankel = np.zeros((n.size, z.size), dtype=complex)
        for idx, order in zip(count(), n):
            hankel[idx, :] = ufunc(order, z, kind)
    else:
        hankel = ufunc(n, z, kind)

    if z.ndim <= 1 or n.ndim <= 1:
        hankel = np.squeeze(hankel)

    return hankel


def _spherical_hankel(n, z, kind):
    if kind == 1:
        hankel = _spspecial.hankel1(n+0.5, z)
    elif kind == 2:
        hankel = _spspecial.hankel2(n+0.5, z)
    hankel = np.sqrt(np.pi/2/z) * hankel

    return hankel


def _spherical_hankel_derivative(n, z, kind):
    return _spherical_hankel(n - 1, z, kind) - (n + 1) / z * _spherical_hankel(
        n, z, kind)


def spherical_harmonic(n, m, theta, phi):
    """The spherical harmonics of order n and degree m.
    The spherical harmonic functions are fully normalized (N3D) and
    include the Condon-Shortley phase according to [#]_.

    n : unsigned int
        The spherical harmonic order
    m : int
        The spherical harmonic degree
    theta : ndarray, double
        The colatitude angle
    phi : ndarray, double
        The azimuth angle

    Returns
    -------
    spherical_harmonic : ndarray, double
        The complex valued spherical harmonic of order n and degree m

    Note
    ----
    This function wraps the spherical harmonic implementation from scipy.

    References
    ----------
    .. [#] B. Rafaely, "Fundamentals of Spherical Array Processing", (2015),
           Springer-Verlag
    """

    return _spspecial.sph_harm_y(n, m, theta, phi)


def spherical_harmonic_real(n, m, theta, phi):
    r"""Real valued spherical harmonic function of order n and degree m
    evaluated at the angles theta and phi.
    The spherical harmonic functions are fully normalized (N3D) and follow
    the AmbiX phase convention, which does not include the Condon-Shortley
    phase [#]_.

    .. math::

        Y_n^m(\theta, \phi) = \sqrt{\frac{2n+1}{4\pi}
        \frac{(n-|m|)!}{(n+|m|)!}} P_n^{|m|}(\cos \theta)
        \begin{cases}
            \displaystyle \cos(|m|\phi),  & \text{if $m \ge 0$} \newline
            \displaystyle \sin(|m|\phi) ,  & \text{if $m < 0$}
        \end{cases}

    References
    ----------
    .. [#]  C. Nachbar, F. Zotter, E. Deleflie, and A. Sontacchi, “Ambix - A
            Suggested Ambisonics Format (revised by F. Zotter),” International
            Symposium on Ambisonics and Spherical Acoustics,
            vol. 3, pp. 1–11, 2011.

    Parameters
    ----------
    n : unsigned int
        The spherical harmonic order
    m : int
        The spherical harmonic degree
    theta : ndarray, double
        The colatitude angle
    phi : ndarray, double
        The azimuth angle

    Returns
    -------
    spherical_harmonic : ndarray, double
        The real valued spherial harmonic of order n and degree m
    """
    # careful here, scipy uses phi as the elevation angle and
    # theta as the azimuth angle
    Y_nm_cplx = spherical_harmonic(n, m, theta, phi)

    if m == 0:
        Y_nm = np.real(Y_nm_cplx)
    elif m > 0:
        Y_nm = np.real(Y_nm_cplx) * np.sqrt(2)
    elif m < 0:
        Y_nm = np.imag(Y_nm_cplx) * np.sqrt(2) * float(-1)**(m+1)

    Y_nm *= float(-1)**(m)

    return Y_nm


def spherical_harmonic_derivative_phi(n, m, theta, phi):
    """Calculate the derivative of the spherical harmonics with respect to
    the azimuth angle phi.

    Parameters
    ----------
    n : int
        Spherical harmonic order
    m : int
        Spherical harmonic degree
    theta : double
        Elevation angle 0 < theta < pi
    phi : double
        Azimuth angle 0 < phi < 2*pi

    Returns
    -------
    sh_diff : complex double
        Spherical harmonic derivative
    """
    if m == 0 or n == 0:
        res = np.zeros(phi.shape, dtype=complex)
    else:
        res = spherical_harmonic(n, m, theta, phi) * 1j * m

    return res


def spherical_harmonic_gradient_phi(n, m, theta, phi):
    """Calculate the derivative of the spherical harmonics with respect to
    the azimuth angle phi divided by sin(theta).

    Parameters
    ----------
    n : int
        Spherical harmonic order
    m : int
        Spherical harmonic degree
    theta : double
        Elevation angle 0 < theta < pi
    phi : double
        Azimuth angle 0 < phi < 2*pi

    Returns
    -------
    sh_diff : complex double
        Spherical harmonic derivative
    """
    if m == 0:
        return np.zeros(theta.shape, dtype=complex)

    factor = np.sqrt((2*n+1)/(2*n-1))/2
    exp_phi = np.exp(1j*phi)
    first = np.sqrt((n+m)*(n+m-1)) * exp_phi * spherical_harmonic(
        n-1, m-1, theta, phi)
    second = np.sqrt((n-m) * (n-m-1)) / exp_phi * spherical_harmonic(
        n-1, m+1, theta, phi)
    Ynm_sin_theta = (-1) * factor * (first + second)

    return Ynm_sin_theta * 1j


def spherical_harmonic_derivative_theta(n, m, theta, phi):
    """Calculate the derivative of the spherical harmonics with respect to
    the elevation angle theta.

    Parameters
    ----------
    n : int
        Spherical harmonic order
    m : int
        Spherical harmonic degree
    theta : double
        Elevation angle 0 < theta < pi
    phi : double
        Azimuth angle 0 < phi < 2*pi

    Returns
    -------
    sh_diff : complex double
        Spherical harmonic derivative
    """
    if n == 0:
        return np.zeros(theta.shape, dtype=complex)

    exp_phi = np.exp(1j*phi)
    first = np.sqrt((n-m+1) * (n+m)) * exp_phi * spherical_harmonic(
        n, m-1, theta, phi)
    second = np.sqrt((n-m) * (n+m+1)) / exp_phi * spherical_harmonic(
        n, m+1, theta, phi)
    return (first-second)/2 * (-1)


def legendre_function(n, m, z, cs_phase=True):
    r"""
    Legendre function of order n and degree m with argument z.

    The Legendre function is defined as ([#]_ Eq. 1.30, 1.33, and 1.34)

    .. math::

        P_n^m(z) = (-1)^m(1-z^2)^{m/2}\frac{d^m}{dz^m}P_n(z),

    where the Condon-Shortley phase term :math:`(-1)^m` is dropped when
    ``cs_phase = False`` and the Legendre Polynomial :math:`P_n(z)`.

    Parameters
    ----------
    n : int
        The order
    m : int
        The degree
    z : ndarray, double
        The argument as an array
    cs_phase : bool, optional
        Whether to include the Condon-Shortley phase term :math:`(-1)^m`
        or not. Default is ``True``.

    Returns
    -------
    legendre : ndarray, double
        The Legendre function. This will return zeros if :math:`|m| > n`.

    References
    ----------
    .. [#] Rafaely, B. (2019). Fundamentals of spherical array processing
           (2nd ed.). Springer. https://doi.org/10.1007/978-3-319-99561-8

    Note
    ----
    This is a wrapper for the Legendre function implementation from scipy. The
    scipy implementation uses the Condon-Shortley phase. Therefore, the sign
    needs to be flipped here for uneven degrees when dropping the
    Condon-Shortley phase.
    """
    z = np.atleast_1d(z)

    # squeeze required because the legendre function introduced in scipy 1.15
    # returns a 2D array, whereas the previous function `lpmn` returned a 1D
    # array
    legendre = np.squeeze(_spspecial.assoc_legendre_p(n, m, z))

    # remove Condon-Shortley phase
    if not cs_phase and m % 2:
        legendre *= -1

    return legendre


def spherical_harmonic_normalization(n, m, norm='full'):
    r"""The normalization factor for real valued spherical harmonics.

    .. math::

        N_n^m = \sqrt{\frac{2n+1}{4\pi}\frac{(n-m)!}{(n+m)!}}

    Parameters
    ----------
    n : int
        The spherical harmonic order.
    m : int
        The spherical harmonic degree.
    norm : 'full', 'semi', optional
        Normalization to use. Can be either fully normalized on the sphere or
        semi-normalized. Default is ``'full'``.

    Returns
    -------
    norm : double
        The normalization factor.
    """
    if np.abs(m) > n:
        factor = 0.0
    elif norm == 'full':
        z = n+m+1
        factor = _spspecial.poch(z, -2*m)
        factor *= (2*n+1)/(4*np.pi)
        if int(m) != 0:
            factor *= 2
        factor = np.sqrt(factor)
    elif norm == 'semi':
        z = n+m+1
        factor = _spspecial.poch(z, -2*m)
        if int(m) != 0:
            factor *= 2
        factor = np.sqrt(factor)
    else:
        raise ValueError("Unknown normalization.")

    return factor


def spherical_harmonic_derivative_theta_real(n, m, theta, phi):
    r"""The derivative of the real valued spherical harmonics with respect
    to the elevation angle :math:`\theta`.

    Parameters
    ----------
    n : int
        The spherical harmonic order.
    m : int
        The spherical harmonic degree.
    theta : ndarray, double
        The elevation angle
    phi : ndarray, double
        The azimuth angle


    Returns
    -------
    derivative : ndarray, double
        The derivative

    Note
    ----
    This implementation does not include the Condon-Shortley phase term.
    """

    m_abs = np.abs(m)

    if n == 0:
        return np.zeros(theta.shape, dtype=float)

    first = (n+m_abs)*(n-m_abs+1) * \
        legendre_function(
            n,
            m_abs-1,
            np.cos(theta),
            cs_phase=False)
    second = legendre_function(
        n,
        m_abs+1,
        np.cos(theta),
        cs_phase=False)
    legendre_diff = 0.5*(first - second)

    N_nm = spherical_harmonic_normalization(n, m_abs)

    phi_term = np.sin(m_abs*phi) if m < 0 else np.cos(m_abs*phi)
    return N_nm * legendre_diff * phi_term


def spherical_harmonic_derivative_phi_real(n, m, theta, phi):
    r"""The derivative of the real valued spherical harmonics with respect
    to the azimuth angle :math:`\phi`.

    Parameters
    ----------
    n : int
        The spherical harmonic order.
    m : int
        The spherical harmonic degree.
    theta : ndarray, double
        The elevation angle
    phi : ndarray, double
        The azimuth angle


    Returns
    -------
    derivative : ndarray, double
        The derivative

    Note
    ----
    This implementation does not include the Condon-Shortley phase term.
    """
    m_abs = np.abs(m)
    if m == 0:
        return np.zeros(theta.shape, dtype=float)

    legendre = legendre_function(n, m_abs, np.cos(theta), cs_phase=False)
    N_nm = spherical_harmonic_normalization(n, m_abs)

    if m < 0:
        phi_term = np.cos(m_abs*phi) * m_abs
    else:
        phi_term = -np.sin(m_abs*phi) * m_abs

    return N_nm * legendre * phi_term


def spherical_harmonic_gradient_phi_real(n, m, theta, phi):
    r"""The gradient of the real valued spherical harmonics with respect
    to the azimuth angle :math:`\phi`.
    The singularities at the poles are avoided using the formulation proposed
    in [#]_.

    Parameters
    ----------
    n : int
        The spherical harmonic order.
    m : int
        The spherical harmonic degree.
    theta : ndarray, double
        The elevation angle
    phi : ndarray, double
        The azimuth angle


    Returns
    -------
    derivative : ndarray, double
        The derivative

    Note
    ----
    This implementation does not include the Condon-Shortley phase term.

    References
    ----------
    .. [#]  J. Du, C. Chen, V. Lesur, and L. Wang, “Non-singular spherical
            harmonic expressions of geomagnetic vector and gradient tensor
            fields in the local north-oriented reference frame,” Geoscientific
            Model Development, vol. 8, no. 7, pp. 1979–1990, Jul. 2015.
    """
    m_abs = np.abs(m)

    if m == 0:
        return np.zeros(theta.shape, dtype=float)

    first = (n+m_abs)*(n+m_abs-1) * \
        legendre_function(
            n-1,
            m_abs-1,
            np.cos(theta),
            cs_phase=False)
    second = legendre_function(
        n-1,
        m_abs+1,
        np.cos(theta),
        cs_phase=False)
    legendre_diff = 0.5*(first + second)
    N_nm = spherical_harmonic_normalization(n, m_abs)

    phi_term = np.cos(m_abs*phi) if m < 0 else -np.sin(m_abs*phi)
    return N_nm * legendre_diff * phi_term


def legendre_coefficients(order):
    """Calculate the coefficients of a Legendre polynomial of the
    specified order.

    Parameters
    ----------
    order : int
        The order of the polynomial

    Returns
    -------
    coefficients : ndarray, double
        The coefficients of the polynomial

    Note
    ----
    This uses :py:mod:`numpy.polynomial.legendre` to compute the coefficients.
    """
    leg = np.polynomial.legendre.Legendre.basis(order)
    return np.polynomial.legendre.leg2poly(leg.coef)


def chebyshev_coefficients(order):
    """Calculate the coefficients of a Chebyshev polynomial of the
    specified order.

    Parameters
    ----------
    order : int
        The order of the polynomial

    Returns
    -------
    coefficients : ndarray, double
        The coefficients of the polynomial

    Note
    ----
    This uses :py:mod:`numpy.polynomial.chebyshev` to compute the coefficients.
    """
    cheb = np.polynomial.chebyshev.Chebyshev.basis(order)
    return np.polynomial.chebyshev.cheb2poly(cheb.coef)



"""Private utility functions for plot module."""

import numpy as np
import matplotlib.pyplot as plt


def _prepare_plot(ax=None, projection=None):
    """
    Returns a figure to plot on.

    Parameters
    ----------
    ax : matplotlib.axes.Axes or list, tuple or ndarray of matplotlib.axes.Axes
        Axes to plot on. The default is None in which case the axes are
        obtained from the current figure. A new figure is created if it does
        not exist.
    projection : str, optional
        Type of projection for the axes. This is only applied if new axes are
        created. Default is ``None`` (2D projection). See
        `matplotlib.projections <https://matplotlib.org/stable/api/projections_api.html>`_
        for more information on projections.

    Returns
    -------
    fig : matplotlib.figure.Figure
        Returns the active figure.
    ax : maptlotlib.axes.Axes
        Returns the current axes.
    """  # noqa: E501
    if ax is None:
        # get current figure or create a new one
        fig = plt.gcf()
        if fig.axes:
            ax = plt.gca()
        else:
            ax = plt.axes(projection=projection)

    else:
        # get figure from axis
        # ax can be array or tuple of two ax objects
        # first axis for the plot, second axis for colorbar placement
        if isinstance(ax, np.ndarray):
            fig = ax.flatten()[0].figure
        elif isinstance(ax, (list, tuple)):
            fig = ax[0].figure
        else:
            fig = ax.figure

    return fig, ax



import pytest
import matplotlib.pyplot as plt
from spharpy.plot._utils import _prepare_plot


@pytest.mark.parametrize(
    ("ax_case", "output_type", "projection"),
    [("none", plt.Axes, "3d"), ("single", plt.Axes, None),
     ("two", list, None)],
)
def test_prepare_plot(ax_case, output_type, projection):
    """
    Test output of :py:func:`~spharpy.plot._utils._prepare_plot`.
    """
    if ax_case == "none":
        input_ax = None
    elif ax_case == "single":
        _, input_ax = plt.subplots()
    else:
        _, axs = plt.subplots(1, 2)
        input_ax = [axs[0], axs[1]]

    fig, ax = _prepare_plot(input_ax, projection)

    assert isinstance(fig, plt.Figure)
    assert isinstance(ax, output_type)

    if isinstance(ax, list):
        assert all(isinstance(ax_, plt.Axes) for ax_ in ax)

    if ax is None and projection is not None:
        assert ax.name == projection



"""
Spherical harmonics and Ambisonics functions. Please refer to the
:doc:`/theory/spherical_harmonic_definition` page for background information
on spherical harmonics.
"""

import numpy as np
import scipy.special as special
import spharpy.special as _special


def acn_to_nm(acn):
    r"""
    Calculate the order n and degree m from the linear coefficient index.

    The linear index corresponds to the Ambisonics Channel Convention [#]_.

    .. math::

        n = \lfloor \sqrt{\mathrm{ACN} + 1} \rfloor - 1

        m = \mathrm{ACN} - n^2 -n

    Parameters
    ----------
    acn : ndarray, int
        Linear index

    Returns
    -------
    n : ndarray, int
        Spherical harmonic order
    m : ndarray, int
        Spherical harmonic degree

    References
    ----------
    .. [#]  C. Nachbar, F. Zotter, E. Deleflie, and A. Sontacchi, “Ambix - A
            Suggested Ambisonics Format (revised by F. Zotter),” International
            Symposium on Ambisonics and Spherical Acoustics,
            vol. 3, pp. 1-11, 2011.
    """
    acn = np.asarray(acn, dtype=int)

    n = (np.ceil(np.sqrt(acn + 1)) - 1)
    m = acn - n**2 - n

    n = n.astype(int, copy=False)
    m = m.astype(int, copy=False)

    return n, m


def nm_to_acn(n, m):
    r"""
    Calculate the linear index coefficient for a order n and degree m.

    The linear index corresponds to the Ambisonics Channel Convention (ACN)
    [#]_.

    .. math::

        \mathrm{ACN} = n^2 + n + m

    References
    ----------
    .. [#]  C. Nachbar, F. Zotter, E. Deleflie, and A. Sontacchi, “Ambix - A
            Suggested Ambisonics Format (revised by F. Zotter),” International
            Symposium on Ambisonics and Spherical Acoustics,
            vol. 3, pp. 1-11, 2011.


    Parameters
    ----------
    n : ndarray, int
        Spherical harmonic order
    m : ndarray, int
        Spherical harmonic degree

    Returns
    -------
    acn : ndarray, int
        Linear index

    """
    n = np.asarray(n, dtype=int)
    m = np.asarray(m, dtype=int)

    if n.size != m.size:
        raise ValueError("n and m need to be of the same size")

    return n**2 + n + m


def nm_to_fuma(n, m):
    r"""
    Calculate the FuMa channel index for a given spherical harmonic order n
    and degree m, according to the FuMa (Furse-Malham)
    Channel Ordering Convention.

    See [#]_ for details.

    Parameters
    ----------
    n : integer, ndarray
        Spherical harmonic order
    m : integer, ndarray
        Spherical harmonic degree

    Returns
    -------
    fuma : integer
        FuMa channel index

    References
    ----------
    .. [#]  D. Malham, "Higher order Ambisonic systems” Space in Music –
             Music in Space (Mphil thesis). University of York. pp. 2–3., 2003.
    """

    fuma_mapping = [0, 2, 3, 1, 8, 6, 4, 5, 7, 15, 13, 11, 9, 10, 12, 14]

    n = np.asarray([n], dtype=int)
    m = np.asarray([m], dtype=int)

    if n.shape != m.shape:
        raise ValueError("n and m need to be of the same size")

    # convert (n, m) to the ACN index
    acn = nm_to_acn(n, m)

    if np.any(acn < 0) or np.any(acn >= len(fuma_mapping)):
        raise ValueError("nm2fuma only supports up to 3rd order")

    acn = np.atleast_2d(acn).T
    fuma = np.array([], dtype=int)
    for a in acn:
        fuma = np.append(fuma, fuma_mapping.index(a))

    return fuma


def fuma_to_nm(fuma):
    r"""
    Calculate the spherical harmonic order n and degree m for a linear
    coefficient index, according to the FuMa (Furse-Malham)
    Channel Ordering Convention.

    See [#]_ for details.

    FuMa = WXYZ | RSTUV | KLMNOPQ
    ACN = WYZX | VTRSU | QOMKLNP

    Parameters
    ----------
    fuma : integer, ndarray
        FuMa channel index

    Returns
    -------
    n : integer, ndarray
        Spherical harmonic order
    m : integer, ndarray
        Spherical harmonic degree

    References
    ----------
    .. [#]  D. Malham, "Higher order Ambisonic systems” Space in Music –
             Music in Space (Mphil thesis). University of York. pp. 2–3., 2003.
    """

    fuma_mapping = [0, 2, 3, 1, 8, 6, 4, 5, 7, 15, 13, 11, 9, 10, 12, 14]

    if not isinstance(fuma, np.ndarray):
        fuma = np.asarray([fuma], dtype=int)

    if np.any(fuma) < 0 or np.any(fuma >= len(fuma_mapping)):
        raise ValueError(
            "Invalid FuMa channel index, must be between 0 and 15 "
            "(supported up to 3rd order)")

    acn = np.array([], dtype=int)
    for f in fuma:
        acn = np.append(acn, fuma_mapping[int(f)])

    n, m = acn_to_nm(acn)
    return n, m


def n3d_to_maxn(acn):
    """
    Calculate the scaling factor which converts from N3D (normalized 3D)
    normalization to max N normalization. ACN must be less or equal to 15.

    Parameters
    ----------
    acn : integer, ndarray
          linear index

    Returns
    -------
    maxN : float
        Scaling factor which converts from N3D to max N
    """

    if not isinstance(acn, np.ndarray):
        acn = np.asarray([acn], dtype=int)

    if np.any(acn) > 15:
        raise ValueError("acn must be less than or "
                         "equal to 15")
    valid_maxN = [
        np.sqrt(1 / 2),
        np.sqrt(1 / 3),
        np.sqrt(1 / 3),
        np.sqrt(1 / 3),
        2 / np.sqrt(15),
        2 / np.sqrt(15),
        np.sqrt(1 / 5),
        2 / np.sqrt(15),
        2 / np.sqrt(15),
        np.sqrt(8 / 35),
        3 / np.sqrt(35),
        np.sqrt(45 / 224),
        np.sqrt(1 / 7),
        np.sqrt(45 / 224),
        3 / np.sqrt(35),
        np.sqrt(8 / 35),
    ]

    maxN = np.array([], dtype=int)
    for a in acn:
        maxN = np.append(maxN, valid_maxN[int(a)])

    return maxN


def n3d_to_sn3d_norm(n):
    """
    Calculate the scaling factor which converts from N3D (normalized 3D)
    normalization to SN3D (Schmidt semi-normalized 3D) normalization.

    Parameters
    ----------
    n : integer, ndarray
        Spherical harmonic order

    Returns
    -------
    sn3d : float, ndarray
        normalization factor which converts from N3D to SN3D
    """
    return 1 / np.sqrt(2 * n + 1)


def renormalize(data, channel_convention, current_norm, target_norm, axis):
    """
    Renormalize spherical harmonics coefficients or basis functions.

    Parameters
    ----------
    data : ndarray
        Data which should be renormalized, either spherical harmonics
        coefficients or basis functions.
    channel_convention : str
        Channel convention of the data which should be renormalized. Valid
        conventions are `"ACN"` or `"FuMa"`.
    current_norm : str
        Current normalization. Valid normalizations are `"N3D"`, `"NM"`,
        `"maxN"`, `"SN3D"`, or `"SNM"`.
    target_norm : str
        Desired normalization. Valid normalizations are `"N3D"`, `"NM"`
        `"maxN"`, `"SN3D"`, or `"SNM"`.
    axis : integer
        Axis along which the renormalization should be applied. The axis
        contains the spherical harmonics coefficients and must hence have
        :math:`Q = (N+1)^2` channels with :math:`N` being the spherical
        harmonics order of data.

    Returns
    -------
    data : ndarray
        Renormalized data
    """
    sh_channels = data.shape[axis]
    if np.sqrt(sh_channels) % 1:
        raise ValueError("Invalid number of SH channels: "
                         f"{data.shape[-2]}. It must match (n_max + 1)^2.")

    if channel_convention not in ["ACN", "FuMa"]:
        raise ValueError("Invalid channel convention. Has to be 'ACN' "
                         f"or 'FuMa', but is {channel_convention}")

    if current_norm not in ["N3D", "NM", "maxN", "SN3D", "SNM"]:
        raise ValueError("Invalid current normalization. Has to be 'N3D', "
                         "'NM', 'maxN', 'SN3D', or 'SNM' "
                         f"but is {current_norm}")

    if target_norm not in ["N3D", "NM", "maxN", "SN3D", "SNM"]:
        raise ValueError("Invalid target normalization. Has to be 'N3D', "
                         "'NM', 'maxN', 'SN3D', or 'SNM' "
                         f"but is {target_norm}")
    if current_norm == target_norm:
        return data

    acn = np.arange(data.shape[axis])

    if channel_convention == "FuMa":
        orders, _ = fuma_to_nm(acn)
    else:
        orders, _ = acn_to_nm(acn)

    # One (re)normalization factor is computed per Ambisonics channel. To make
    # sure that the factors can be applied, new axes must be added. This is
    # done by reshaping to the following shape
    shape = [1] * data.ndim
    shape[axis] = data.shape[axis]

    data_renorm = data.copy()
    # normalize to 'n3d'
    if current_norm == 'NM':
        data_renorm /= np.sqrt(4*np.pi)
    if current_norm == 'SN3D':
        data_renorm /= n3d_to_sn3d_norm(orders).reshape(shape)
    if current_norm == 'SNM':
        data_renorm /= n3d_to_sn3d_norm(orders).reshape(shape)
        data_renorm /= np.sqrt(4*np.pi)
    if current_norm == 'maxN':
        data_renorm /= n3d_to_maxn(acn).reshape(shape)

    # convert to target norm
    if target_norm == "NM":
        data_renorm *= np.sqrt(4*np.pi)
    if target_norm == "SN3D":
        data_renorm *= n3d_to_sn3d_norm(orders).reshape(shape)
    if target_norm == "SNM":
        data_renorm *= n3d_to_sn3d_norm(orders).reshape(shape)
        data_renorm *= np.sqrt(4*np.pi)
    if target_norm == 'maxN':
        data_renorm *= n3d_to_maxn(acn).reshape(shape)

    return data_renorm


def change_channel_convention(data, current, target, axis):
    """
    Change the channel convention of spherical harmonics coefficients
    or basis functions.

    Parameters
    ----------
    data : ndarray
        Data of which channel convention should be changed. Either spherical
        harmonics coefficients or bases functions.
    current : str
        Current channel convention. Valid conventions are `"ACN"` or `"FuMa"`.
    target : str
        Desired channel convention. Valid conventions are `"ACN"` or `"FuMa"`.
    axis : integer
        Axis along which the channel convention should be changed

    Returns
    -------
    data : ndarray
        Data with changed channel convention
    """
    if current not in ["ACN", "FuMa"]:
        raise ValueError("Invalid current channel convention. Has to be "
                         f"'ACN' or 'FuMa', but is {current}")

    if target not in ["ACN", "FuMa"]:
        raise ValueError("Invalid target channel convention. Has to be 'ACN' "
                         f"or 'FuMa', but is {target}")

    if current == target:
        return data

    acn = np.arange(data.shape[axis])
    if current == 'ACN':
        n, m = acn_to_nm(acn)
        idx = nm_to_fuma(n, m)
    elif current == 'FuMa':
        n, m = fuma_to_nm(acn)
        idx = nm_to_acn(n, m)

    return np.take(data, idx, axis=axis)


def spherical_harmonic_basis(
        n_max, coordinates, normalization="N3D", channel_convention="ACN",
        condon_shortley='auto'):
    r"""
    Calculates the complex valued spherical harmonic basis matrix.

    See [#]_ and [#]_ for details.
    See also :py:func:`spherical_harmonic_basis_real`.

    .. math::
        Y_n^m(\theta, \phi) =  CS_m N_{nm} P_{nm}(cos(\theta)) e^{im\phi}

    where:

    - :math:`n` is the degree
    - :math:`m` is the order
    - :math:`P_{nm}` is the associated Legendre function
    - :math:`N_{nm}` is the normalization term
    - :math:`CS_m` is the Condon-Shortley phase term
    - :math:`\theta` is the colatitude (angle from the positive z-axis)
    - :math:`\phi` is the azimuth (angle from the positive x-axis in the
      xy-plane), see :py:mod:`~pyfar.classes.coordinates`

    References
    ----------
    .. [#]  E. G. Williams, Fourier Acoustics. Academic Press, 1999.
    .. [#]  B. Rafaely, Fundamentals of Spherical Array Processing, vol. 8.
            Springer, 2015.

    Parameters
    ----------
    n_max : integer
        Spherical harmonic order
    coordinates : :py:class:`pyfar.Coordinates`, :py:class:`spharpy.SamplingSphere`
        objects with sampling points for which the basis matrix is calculated
    normalization : str, optional
        Normalization convention, either ``'N3D'``, ``'NM'``, ``'maxN'``,
        ``'SN3D'``, or ``'SNM'``.
        (maxN is only supported up to 3rd order)
    channel_convention : str, optional
        Channel ordering convention, either ``'ACN'`` or ``'FuMa'``.
        The default is ``'ACN'``.
        (FuMa is only supported up to 3rd order)
    condon_shortley : bool or str, optional
        Whether to include the Condon-Shortley phase term. If ``True`` or
        ``'auto'``, Condon-Shortley is included, if ``False`` it is not
        included. The default is ``'auto'``.

    Returns
    -------
    Y : ndarray, complex
        Complex spherical harmonic basis matrix

    Examples
    --------
    >>> import spharpy
    >>> n_max = 2
    >>> coordinates = spharpy.samplings.icosahedron()
    >>> Y = spharpy.spherical.spherical_harmonic_basis(n_max, coordinates)

    """  # noqa: E501
    if channel_convention == "FuMa" and n_max > 3:
        raise ValueError(
            "FuMa channel convention is only supported up to 3rd order.")

    if normalization == "maxN" and n_max > 3:
        raise ValueError(
            "MaxN normalization is only supported up to 3rd order.")

    if not isinstance(condon_shortley, bool) and condon_shortley != 'auto':
        raise ValueError(
            "Condon_shortley has to be a bool, or 'auto'.")

    if condon_shortley == 'auto':
        condon_shortley = True

    n_coeff = (n_max + 1) ** 2

    basis = np.zeros((coordinates.csize, n_coeff), dtype=complex)

    for acn in range(n_coeff):
        if channel_convention == "FuMa":
            order, degree = fuma_to_nm(acn)
        else:
            order, degree = acn_to_nm(acn)
        basis[:, acn] = _special.spherical_harmonic(
            order, degree, coordinates.colatitude, coordinates.azimuth,
        )
        if normalization == "NM":
            basis[:, acn] *= np.sqrt(4*np.pi)
        if normalization == "SN3D":
            basis[:, acn] *= n3d_to_sn3d_norm(order)
        if normalization == "SNM":
            basis[:, acn] *= n3d_to_sn3d_norm(order) * np.sqrt(4*np.pi)
        elif normalization == "maxN":
            basis[:, acn] *= n3d_to_maxn(acn)
        if not condon_shortley:
            # Condon-Shortley phase term is already included in
            # the special.spherical_harmonic function
            # so need to divide by (-1)^m
            basis[:, acn] /= (-1) ** float(degree.squeeze())
    return basis


def spherical_harmonic_basis_gradient(n_max, coordinates, normalization="N3D",
                                      channel_convention="ACN",
                                      condon_shortley='auto'):
    r"""
    Calculates the unit sphere gradients of the complex spherical harmonics.

    This implementation avoids singularities at the poles using identities
    derived in [#]_.

    The angular parts of the gradient are defined as

    .. math::

        \nabla_{(\theta, \phi)} Y_n^m(\theta, \phi) =
        \frac{1}{\sin \theta} \frac{\partial Y_n^m(\theta, \phi)}
        {\partial \phi} \vec{e}_\phi +
        \frac{\partial Y_n^m(\theta, \theta)}
        {\partial \theta} \vec{e}_\theta .


    References
    ----------
    .. [#]  J. Du, C. Chen, V. Lesur, and L. Wang, “Non-singular spherical
            harmonic expressions of geomagnetic vector and gradient tensor
            fields in the local north-oriented reference frame,” Geoscientific
            Model Development, vol. 8, no. 7, pp. 1979-1990, Jul. 2015.

    Parameters
    ----------
    n_max : int
        Spherical harmonic order
    coordinates : :py:class:`pyfar.Coordinates`, :py:class:`spharpy.SamplingSphere`
        objects with sampling points for which the basis matrix is
        calculated
    normalization : str, optional
        Normalization convention, either ``'N3D'``, ``'NM'``, ``'maxN'``,
        ``'SN3D'``, or ``'SNM'``.
        (maxN is only supported up to 3rd order)
    channel_convention : str, optional
        Channel ordering convention, either ``'acn'`` or ``'fuma'``.
        The default is ``'acn'``.
        (FuMa is only supported up to 3rd order)
    condon_shortley : bool or str, optional
        Whether to include the Condon-Shortley phase term. If ``True`` or
        ``'auto'``, Condon-Shortley is included, if ``False`` it is not
        included. The default is ``'auto'``.

    Returns
    -------
    grad_theta : ndarray, complex
        Gradient with regard to the co-latitude angle.
    grad_azimuth : ndarray, complex
        Gradient with regard to the azimuth angle.

    Examples
    --------
    >>> import spharpy
    >>> n_max = 2
    >>> coordinates = spharpy.samplings.icosahedron()
    >>> grad_theta, grad_phi = /
        spharpy.spherical.spherical_harmonic_basis_gradient(n_max, coordinates)

    """  # noqa: E501
    if channel_convention == "FuMa" and n_max > 3:
        raise ValueError(
            "FuMa channel convention is only supported up to 3rd order.")

    if normalization == "maxN" and n_max > 3:
        raise ValueError(
            "MaxN normalization is only supported up to 3rd order.")

    if not isinstance(condon_shortley, bool) and condon_shortley != 'auto':
        raise ValueError(
            "Condon_shortley has to be a bool, or 'auto'.")

    if condon_shortley == 'auto':
        condon_shortley = True

    n_points = coordinates.csize
    n_coeff = (n_max+1)**2
    theta = coordinates.colatitude
    phi = coordinates.azimuth
    grad_theta = np.zeros((n_points, n_coeff), dtype=complex)
    grad_phi = np.zeros((n_points, n_coeff), dtype=complex)

    for acn in range(n_coeff):
        if channel_convention == "FuMa":
            n, m = fuma_to_nm(acn)
        else:
            n, m = acn_to_nm(acn)

        grad_theta[:, acn] = _special.spherical_harmonic_derivative_theta(
            n, m, theta, phi)
        grad_phi[:, acn] = _special.spherical_harmonic_gradient_phi(
            n, m, theta, phi)

        factor = 1.
        if normalization == "NM":
            factor = np.sqrt(4*np.pi)
        if normalization == "SN3D":
            factor = n3d_to_sn3d_norm(n)
        if normalization == "SNM":
            factor = n3d_to_sn3d_norm(n) * np.sqrt(4*np.pi)
        elif normalization == "maxN":
            factor *= n3d_to_maxn(acn)

        if not condon_shortley:
            # Condon-Shortley phase term is already included in
            # the special.spherical_harmonic function
            # so need to divide by (-1)^m
            factor /= (-1) ** float(m)

        grad_theta[:, acn] *= factor
        grad_phi[:, acn] *= factor

    return grad_theta, grad_phi


def spherical_harmonic_basis_real(
        n_max, coordinates, normalization="N3D", channel_convention="ACN",
        condon_shortley='auto'):
    r"""
    Calculates the real valued spherical harmonic basis matrix.
    See also :py:func:`spherical_harmonic_basis`.

    .. math::

        Y_n^m(\theta, \phi) = \sqrt{\frac{2n+1}{4\pi}
        \frac{(n-|m|)!}{(n+|m|)!}} P_n^{|m|}(\cos \theta)
        \begin{cases}
            \displaystyle \cos(|m|\phi),  & \text{if $m \ge 0$} \newline
            \displaystyle \sin(|m|\phi) ,  & \text{if $m < 0$}
        \end{cases}

    Parameters
    ----------
    n_max : int
        Spherical harmonic order
    coordinates : :py:class:`pyfar.Coordinates`, :py:class:`spharpy.SamplingSphere`
        objects with sampling points for which the basis matrix is
        calculated
    normalization : str, optional
        Normalization convention, either ``'N3D'``, ``'NM'``, ``'maxN'``,
        ``'SN3D'``, or ``'SNM'``.
        (maxN is only supported up to 3rd order)
    channel_convention : str, optional
        Channel ordering convention, either ``'ACN'`` or ``'FuMa'``.
        The default is ``'ACN'``.
        (FuMa is only supported up to 3rd order)
    condon_shortley : bool or str, optional
        Whether to include the Condon-Shortley phase term. If ``True``,
        Condon-Shortley is included, if ``False`` or ``'auto'``, it is not
        included. The default is ``'auto'``.

    Returns
    -------
    Y : ndarray, float
        Real valued spherical harmonic basis matrix.

    """  # noqa: E501
    if channel_convention == "FuMa" and n_max > 3:
        raise ValueError(
            "FuMa channel convention is only supported up to 3rd order.")

    if normalization == "maxN" and n_max > 3:
        raise ValueError(
            "MaxN normalization is only supported up to 3rd order.")

    if not isinstance(condon_shortley, bool) and condon_shortley != 'auto':
        raise ValueError(
            "Condon_shortley has to be a bool, or 'auto'.")

    if condon_shortley == 'auto':
        condon_shortley = False

    n_coeff = (n_max + 1) ** 2

    basis = np.zeros((coordinates.csize, n_coeff), dtype=float)

    for acn in range(n_coeff):
        if channel_convention == "FuMa":
            order, degree = fuma_to_nm(acn)
        else:
            order, degree = acn_to_nm(acn)
        basis[:, acn] = _special.spherical_harmonic_real(
            order, degree, coordinates.colatitude, coordinates.azimuth,
        )
        if normalization == "NM":
            basis[:, acn] *= np.sqrt(4*np.pi)
        if normalization == "SN3D":
            basis[:, acn] *= n3d_to_sn3d_norm(order)
        if normalization == "SNM":
            basis[:, acn] *= n3d_to_sn3d_norm(order) * np.sqrt(4*np.pi)
        elif normalization == "maxN":
            basis[:, acn] *= n3d_to_maxn(acn)
        if condon_shortley:
            # Condon-Shortley phase term is not included in
            # the special.spherical_harmonic function
            basis[:, acn] *= (-1) ** float(degree.squeeze())

    return basis


def spherical_harmonic_basis_gradient_real(n_max, coordinates,
                                           normalization="N3D",
                                           channel_convention="ACN",
                                           condon_shortley='auto'):
    r"""
    Calculates the unit sphere gradients of the real valued spherical
    harmonics.

    This implementation avoids singularities at the poles using identities
    derived in [#]_.

    The angular parts of the gradient are defined as

    .. math::

        \nabla_{(\theta, \phi)} Y_n^m(\theta, \phi) =
        \frac{1}{\sin \theta} \frac{\partial Y_n^m(\theta, \phi)}
        {\partial \phi} \vec{e}_\phi +
        \frac{\partial Y_n^m(\theta, \theta)}
        {\partial \theta} \vec{e}_\theta .


    References
    ----------
    .. [#]  J. Du, C. Chen, V. Lesur, and L. Wang, “Non-singular spherical
            harmonic expressions of geomagnetic vector and gradient tensor
            fields in the local north-oriented reference frame,” Geoscientific
            Model Development, vol. 8, no. 7, pp. 1979-1990, Jul. 2015.

    Parameters
    ----------
    n_max : int
        Spherical harmonic order
    coordinates : :py:class:`pyfar.Coordinates`, :py:class:`spharpy.SamplingSphere`
        objects with sampling points for which the basis matrix is
        calculated
    normalization : str, optional
        Normalization convention, either ``'N3D'``, ``'NM'``,
        ``'maxN'``, ``'SN3D'``, or ``'SNM'``.
        (maxN is only supported up to 3rd order)
    channel_convention : str, optional
        Channel ordering convention, either ``'ACN'`` or ``'FuMa'``.
        The default is ``'ACN'``.
        (FuMa is only supported up to 3rd order)
    condon_shortley : bool or str, optional
        Whether to include the Condon-Shortley phase term. If ``True``,
        Condon-Shortley is included, if ``False`` or ``'auto'``, it is not
        included. The default is ``'auto'``.

    Returns
    -------
    grad_theta : ndarray, float
        Gradient with respect to the co-latitude angle.
    grad_phi : ndarray, float
        Gradient with respect to the azimuth angle.

    """  # noqa: E501
    if channel_convention == "FuMa" and n_max > 3:
        raise ValueError(
            "FuMa channel convention is only supported up to 3rd order.")

    if normalization == "maxN" and n_max > 3:
        raise ValueError(
            "MaxN normalization is only supported up to 3rd order.")

    if not isinstance(condon_shortley, bool) and condon_shortley != 'auto':
        raise ValueError(
            "Condon_shortley has to be a bool, or 'auto'.")

    if condon_shortley == 'auto':
        condon_shortley = False

    n_points = coordinates.csize
    n_coeff = (n_max + 1) ** 2
    theta = coordinates.colatitude
    phi = coordinates.azimuth
    grad_theta = np.zeros((n_points, n_coeff), dtype=float)
    grad_phi = np.zeros((n_points, n_coeff), dtype=float)

    for acn in range(n_coeff):
        if channel_convention == "FuMa":
            n, m = fuma_to_nm(acn)
        else:
            n, m = acn_to_nm(acn)

        grad_theta[:, acn] = \
            _special.spherical_harmonic_derivative_theta_real(
                n, m, theta, phi)
        grad_phi[:, acn] = \
            _special.spherical_harmonic_gradient_phi_real(
                n, m, theta, phi)

        factor = 1.0
        if normalization == "NM":
            factor = np.sqrt(4*np.pi)
        if normalization == "SN3D":
            factor = n3d_to_sn3d_norm(n)
        if normalization == "SNM":
            factor = n3d_to_sn3d_norm(n) * np.sqrt(4*np.pi)
        elif normalization == "maxN":
            factor *= n3d_to_maxn(acn)

        if condon_shortley:
            # Condon-Shortley phase term is not included in
            # the special.spherical_harmonic function
            factor *= (-1) ** float(m)

        grad_theta[:, acn] *= factor
        grad_phi[:, acn] *= factor

    return grad_theta, grad_phi


def modal_strength(n_max,
                   kr,
                   arraytype='rigid'):
    r"""
    Modal strength function for microphone arrays.

    .. math::

        b(kr) =
        \begin{cases}
            \displaystyle 4\pi i^n j_n(kr),  & \text{open} \newline
            \displaystyle  4\pi i^{(n-1)} \frac{1}{(kr)^2 h_n^\prime(kr)},
                & \text{rigid} \newline
            \displaystyle  4\pi i^n (j_n(kr) - i j_n^\prime(kr)),
                & \text{cardioid}
        \end{cases}


    This implementation uses the second order Hankel function, see [#]_ for an
    overview of the corresponding sign conventions.

    References
    ----------
    .. [#]  V. Tourbabin and B. Rafaely, “On the Consistent Use of Space and
            Time Conventions in Array Processing,” vol. 101, pp. 470–473, 2015.


    Parameters
    ----------
    n_max : int
        The spherical harmonic order
    kr : ndarray, float
        Wave number * radius
    arraytype : str
        Array configuration. Can be omnidirectional microphones mounted on a
        ``'rigid'`` sphere or on a virtual ``'open'`` sphere, or ``'cardioid'``
        microphones on an open sphere. Default is ``'rigid'``.

    Returns
    -------
    B : ndarray, float
        Modal strength diagonal matrix

    """
    n_coeff = (n_max+1)**2
    n_bins = kr.shape[0]

    modal_strength_mat = np.zeros((n_bins, n_coeff, n_coeff), dtype=complex)

    for n in range(n_max+1):
        bn = _modal_strength(n, kr, arraytype)
        for m in range(-n, n+1):
            acn = n*n + n + m
            modal_strength_mat[:, acn, acn] = bn

    return np.squeeze(modal_strength_mat)


def _modal_strength(n, kr, config):
    """Helper function for the calculation of the modal strength for
    plane waves.
    """
    if config == 'open':
        ms = 4*np.pi*pow(1.0j, n) * _special.spherical_bessel(n, kr)
    elif config == 'rigid':
        ms = 4*np.pi*pow(1.0j, n+1) / \
            _special.spherical_hankel(n, kr, derivative=True) / (kr)**2
    elif config == 'cardioid':
        ms = 4*np.pi*pow(1.0j, n) * \
            (_special.spherical_bessel(n, kr) -
                1.0j * _special.spherical_bessel(n, kr, derivative=True))
    else:
        raise ValueError("Invalid configuration.")

    return ms


def aperture_vibrating_spherical_cap(
        n_max,
        rad_sphere,
        rad_cap):
    r"""
    Aperture function for a vibrating spherical cap.

    The cap has radius :math:`r_c` and is mounted in a rigid sphere with
    radius :math:`r_s` [#]_, [#]_

    .. math::

        a_n (r_{s}, \alpha) = 4 \pi
        \begin{cases}
            \displaystyle \left(2n+1\right)\left[
                P_{n-1} \left(\cos\alpha\right) -
                P_{n+1} \left(\cos\alpha\right) \right],
                & {n>0} \newline
            \displaystyle  (1 - \cos\alpha)/2,  & {n=0}
        \end{cases}

    where :math:`\alpha = \arcsin \left(\frac{r_c}{r_s} \right)` is the
    aperture angle.

    Parameters
    ----------
    n_max : integer, ndarray
        Maximal spherical harmonic order
    rad_sphere : double, ndarray
        Radius of the sphere
    rad_cap : double
        Radius of the vibrating cap

    Returns
    -------
    A : ndarray, float
        Aperture function in diagonal matrix form with shape
        :math:`[(n_{max}+1)^2~\times~(n_{max}+1)^2]`

    References
    ----------
    .. [#]  E. G. Williams, Fourier Acoustics. Academic Press, 1999.
    .. [#]  B. Rafaely and D. Khaykin, “Optimal Model-Based Beamforming and
             Independent Steering for Spherical Loudspeaker Arrays,” IEEE
             Transactions on Audio, Speech, and Language Processing, vol. 19,
             no. 7, pp. 2234-2238, 2011

    Notes
    -----
    Eq. (3) in the second Ref. contains an error, here, the power of 2 on pi is
    omitted on the normalization term.

    """
    angle_cap = np.arcsin(rad_cap / rad_sphere)
    arg = np.cos(angle_cap)
    n_sh = (n_max+1)**2

    aperture = np.zeros((n_sh, n_sh), dtype=float)

    aperture[0, 0] = (1-arg)*2*np.pi
    for n in range(1, n_max+1):
        legendre_minus = special.legendre(n-1)(arg)
        legendre_plus = special.legendre(n+1)(arg)
        legendre_term = legendre_minus - legendre_plus
        for m in range(-n, n+1):
            acn = nm_to_acn(n, m)
            aperture[acn, acn] = legendre_term * 4 * np.pi / (2*n+1)

    return aperture


def radiation_from_sphere(
        n_max,
        rad_sphere,
        k,
        distance,
        density_medium=1.2,
        speed_of_sound=343.0):
    r"""
    Radiation function in SH for a vibrating spherical cap.
    Includes the radiation impedance and the propagation to a arbitrary
    distance from the sphere.
    The sign and phase conventions result in a positive pressure response for
    a positive cap velocity with the intensity vector pointing away from the
    source [#]_, [#]_.

    TODO: This function does not have a test yet.

    References
    ----------
    .. [#]  E. G. Williams, Fourier Acoustics. Academic Press, 1999.
    .. [#]  F. Zotter, A. Sontacchi, and R. Höldrich, “Modeling a spherical
            loudspeaker system as multipole source,” in Proceedings of the 33rd
            DAGA German Annual Conference on Acoustics, 2007, pp. 221-222.


    Parameters
    ----------
    n_max : integer
        Maximal spherical harmonic order
    rad_sphere : float
        Radius of the sphere
    k : ndarray, float
        Wave number
    distance : float
        Radial distance from the center of the sphere
    density_medium : float
        Density of the medium surrounding the sphere. Default is ``1.2``.
        for air.
    speed_of_sound : float
        Speed of sound in m/s. Default is ``343.0``.

    Returns
    -------
    R : ndarray, float
        Radiation function in diagonal matrix form with shape
        :math:`[K \times (n_{max}+1)^2~\times~(n_{max}+1)^2]`

    """
    n_sh = (n_max+1)**2

    k = np.atleast_1d(k)
    n_bins = k.shape[0]
    radiation = np.zeros((n_bins, n_sh, n_sh), dtype=complex)

    for n in range(n_max+1):
        hankel = _special.spherical_hankel(n, k*distance, kind=2)
        hankel_prime = _special.spherical_hankel(
            n, k*rad_sphere, kind=2, derivative=True)
        radiation_order = -1j * hankel/hankel_prime * \
            density_medium * speed_of_sound
        for m in range(-n, n+1):
            acn = nm_to_acn(n, m)
            radiation[:, acn, acn] = radiation_order

    return radiation


def sid(n_max):
    """Calculates the SID indices up to spherical harmonic order n_max.
    The SID indices were originally proposed by Daniel [#]_, more recently
    ACN indexing has been favored and is used in the AmbiX format [#]_.

    Parameters
    ----------
    n_max : int
        The maximum spherical harmonic order

    Returns
    -------
    sid_n : ndarray, int
        The SID indices for all orders
    sid_m : ndarray, int
        The SID indices for all degrees

    References
    ----------
    .. [#]  J. Daniel, “Représentation de champs acoustiques, application à la
            transmission et à la reproduction de scènes sonores complexes dans
            un contexte multimédia,” Dissertation, l’Université Paris 6, Paris,
            2001.

    .. [#]  C. Nachbar, F. Zotter, E. Deleflie, and A. Sontacchi, “Ambix - A
            Suggested Ambisonics Format (revised by F. Zotter),” International
            Symposium on Ambisonics and Spherical Acoustics,
            vol. 3, pp. 1-11, 2011.

    """
    n_sh = (n_max+1)**2
    sid_n = sph_identity_matrix(n_max, 'n-nm').T @ np.arange(0, n_max+1)
    sid_m = np.zeros(n_sh, dtype=int)
    idx_n = 0
    for n in range(1, n_max+1):
        for m in range(1, n+1):
            sid_m[idx_n + 2*m-1] = n-m+1
            sid_m[idx_n + 2*m] = -(n-m+1)
        sid_m[idx_n + 2*n + 1] = 0
        idx_n += 2*n+1

    return sid_n, sid_m


def sid_to_acn(n_max):
    """Convert from SID channel indexing to ACN indeces.
    Returns the indices to achieve a corresponding linear ACN indexing.

    Parameters
    ----------
    n_max : int
        The maximum spherical harmonic order.

    Returns
    -------
    acn : ndarray, int
        The SID indices sorted according to a respective linear ACN indexing.
    """
    sid_n, sid_m = sid(n_max)
    linear_sid = nm_to_acn(sid_n, sid_m)
    return np.argsort(linear_sid)


def sph_identity_matrix(n_max, matrix_type='n-nm'):
    """Calculate a spherical harmonic identity matrix.

    Parameters
    ----------
    n_max : int
        The spherical harmonic order.
    matrix_type : str, optional
        The type of identity matrix. Currently only ``'n-nm'`` is implemented.

    Returns
    -------
    identity_matrix : ndarray, int
        The spherical harmonic identity matrix.

    Examples
    --------
    The identity matrix can for example be used to decompress from order only
    vectors to a full order and degree representation.

    >>> import spharpy
    >>> import matplotlib.pyplot as plt
    >>> n_max = 2
    >>> E = spharpy.spherical.sph_identity_matrix(n_max, matrix_type='n-nm')
    >>> a_n = [1, 2, 3]
    >>> a_nm = E.T @ a_n
    >>> a_nm
    array([1, 2, 2, 2, 3, 3, 3, 3, 3])

    The matrix E in this case has the following form.

    .. plot::

        >>> import spharpy
        >>> import matplotlib.pyplot as plt
        >>> n_max = 2
        >>> E = spharpy.spherical.sph_identity_matrix(n_max, matrix_type='n-nm')
        >>> plt.matshow(E, cmap=plt.get_cmap('Greys'))
        >>> plt.gca().set_aspect('equal')

    """  # noqa: E501
    n_sh = (n_max+1)**2

    if matrix_type != 'n-nm':
        raise NotImplementedError

    identity_matrix = np.zeros((n_max+1, n_sh), dtype=int)

    for n in range(n_max+1):
        m = np.arange(-n, n+1)
        linear_nm = nm_to_acn(np.tile(n, m.shape), m)
        identity_matrix[n, linear_nm] = 1

    return identity_matrix



import spharpy.spherical as sh
import numpy as np
import pytest


@pytest.mark.parametrize(('index', 'expected'), [
    (0, np.sqrt(1/2)), (1, np.sqrt(1/3)), (2, np.sqrt(1/3))])
def test_n3d_to_maxn_single_val(index, expected):
    assert sh.n3d_to_maxn(index) == expected


def test_n3d_to_maxn_array():
    acn = np.array([0, 1, 2])
    maxN_ref = np.array([np.sqrt(1 / 2),
                         np.sqrt(1 / 3),
                         np.sqrt(1 / 3)])

    maxN_norm = sh.n3d_to_maxn(acn)
    np.testing.assert_equal(maxN_norm, maxN_ref)


@pytest.mark.parametrize(('index', 'expected'), [
    (0, 1), (1, 1/np.sqrt(3)), (2, 1/np.sqrt(5))])
def test_n3d_to_sn3d_norm_single_val(index, expected):
    assert sh.n3d_to_sn3d_norm(index) == expected


def test_n3d_to_sn3d_norm_array():
    n = np.array([0, 1, 2])
    sn3d_norm_ref = np.array([1, 1/np.sqrt(3), 1/np.sqrt(5)])
    sn3d_norm = sh.n3d_to_sn3d_norm(n)

    np.testing.assert_equal(sn3d_norm, sn3d_norm_ref)



import pytest
from spharpy.classes import SphericalHarmonicSignal
from spharpy.classes import SphericalHarmonicDefinition
import numpy as np
import re


def test_spherical_harmonic_signal_init():
    """Test init SphericalHarmonicsSignal."""

    data = np.array([[1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.]]).reshape(1, 4, 3)
    signal = SphericalHarmonicSignal(data,
                                     44100, basis_type='real',
                                     channel_convention='ACN',
                                     normalization='N3D',
                                     condon_shortley=False)
    assert isinstance(signal, SphericalHarmonicSignal)


def test_sh_signal_from_sh_definition():
    """Test init SphericalHarmonicsSignal."""

    shd = SphericalHarmonicDefinition(
        n_max=0, basis_type="real",
        channel_convention="ACN",
        normalization="N3D",
        condon_shortley=False)

    data = np.array([[1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.]]).reshape(1, 4, 3)

    signal_def = SphericalHarmonicSignal.from_definition(
        sh_definition=shd, data=data, sampling_rate=44100)

    assert isinstance(signal_def, SphericalHarmonicSignal)

    signal = SphericalHarmonicSignal(data,
                                     44100, basis_type='real',
                                     channel_convention='ACN',
                                     normalization='N3D',
                                     condon_shortley=False)
    assert signal_def == signal


def test_spherical_harmonic_signal_init_condon_shortley():
    """
    Test if Condon-Shortley is set properly in init
    SphericalHarmonicsSignal.
    """

    data = np.array([[1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.]]).reshape(1, 4, 3)
    signal = SphericalHarmonicSignal(data,
                                     44100, basis_type='real',
                                     channel_convention='ACN',
                                     condon_shortley=False,
                                     normalization='N3D')
    assert not signal.condon_shortley

    signal = SphericalHarmonicSignal(data.astype(np.complex128),
                                     44100, basis_type='complex',
                                     channel_convention='ACN',
                                     normalization='N3D',
                                     condon_shortley=True,
                                     is_complex=True)
    assert signal.condon_shortley


def test_spherical_harmonic_signal_wrong_dimensions():
    """Test dimensions of SH coefficient data."""

    # test if sh channels are valid
    data = np.array([[1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.]]).reshape(1, 5, 3)

    with pytest.raises(ValueError,
                   match=re.escape("Invalid number of spherical harmonic "
                                   f"channels: {data.shape[-2]}. It must "
                                   "match (n_max + 1)^2.")):
        SphericalHarmonicSignal(data,
                                44100, basis_type='real',
                                channel_convention='ACN',
                                condon_shortley=False,
                                normalization='N3D')


def test_spherical_harmonic_signal_init_multichannel():
    """Test init SphercalHarmonicsSignal."""
    sh_coeffs = np.zeros((2, 4, 16))
    signal = SphericalHarmonicSignal(sh_coeffs,
                                     44100, basis_type='real',
                                     channel_convention='ACN',
                                     normalization='N3D',
                                     condon_shortley=False)
    assert isinstance(signal, SphericalHarmonicSignal)


def test_nmax_getter():
    """Test nmax getter."""
    data = np.array([[1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.]]).reshape(1, 4, 3)
    signal = SphericalHarmonicSignal(data,
                                     44100, basis_type='real',
                                     channel_convention='ACN',
                                     normalization='N3D',
                                     condon_shortley=False)
    assert signal.n_max == 1
    assert isinstance(signal.n_max, int)


def test_init_wrong_basis_type():
    data = np.array([[1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.]]).reshape(1, 4, 3)
    with pytest.raises(ValueError,
                match="Invalid basis type, only "
                      "'complex' and 'real' are supported"):
        SphericalHarmonicSignal(data,
                                44100, basis_type='invalid_basis_type',
                                channel_convention='ACN',
                                normalization='N3D',
                                condon_shortley=False)


def test_basis_type_getter():
    """Test basis_type getter."""
    data = np.array([[1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.]]).reshape(1, 4, 3)
    signal = SphericalHarmonicSignal(data,
                                     44100, basis_type='real',
                                     channel_convention='ACN',
                                     normalization='N3D',
                                     condon_shortley=False)
    assert signal.basis_type == 'real'


def test_init_wrong_normalization():
    data = np.array([[1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.]]).reshape(1, 4, 3)

    with pytest.raises(ValueError,
                       match="Invalid normalization, currently only 'N3D', "
                             "'NM', 'maxN', 'SN3D', 'SNM' are supported"):
        SphericalHarmonicSignal(data,
                                44100, basis_type='real',
                                channel_convention='ACN',
                                normalization='invalid_normalization',
                                condon_shortley=False)


def test_spherical_harmonic_signal_normalization_setter():
    data = np.array([[1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.]]).reshape(1, 4, 3)
    signal = SphericalHarmonicSignal(data,
                                     44100, basis_type='real',
                                     channel_convention='ACN',
                                     normalization='N3D',
                                     condon_shortley=False)
    signal.normalization = 'SN3D'
    assert signal.normalization == 'SN3D'

    signal.normalization = 'SNM'
    assert signal.normalization == 'SNM'

    signal.normalization = 'maxN'
    assert signal.normalization == 'maxN'

    signal.normalization = 'N3D'
    assert signal.normalization == 'N3D'

    signal.normalization = 'NM'
    assert signal.normalization == 'NM'


def test_init_wrong_channel_convention():
    data = np.array([[1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.]]).reshape(1, 4, 3)

    with pytest.raises(
        ValueError,
        match="Invalid channel convention, currently only 'ACN' "
              "and 'FuMa' are supported"):
        SphericalHarmonicSignal(data,
                                44100, basis_type='real',
                                channel_convention='invalid_convention',
                                normalization='N3D',
                                condon_shortley=False)


def test_spherical_harmonic_signal_channel_convention_setter():
    data = np.array([[1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.]]).reshape(1, 4, 3)
    signal = SphericalHarmonicSignal(data,
                                     44100, basis_type='real',
                                     channel_convention='ACN',
                                     normalization='N3D',
                                     condon_shortley=False)
    signal.channel_convention = 'FuMa'
    assert signal.channel_convention == 'FuMa'

    signal.channel_convention = 'ACN'
    assert signal.channel_convention == 'ACN'


def test_spherical_harmonic_signal_change_channel_convention():
    data = np.array([[1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.],
                     [1., 2., 3.]]).reshape(1, 4, 3)
    signal = SphericalHarmonicSignal(data,
                                     44100, basis_type='real',
                                     channel_convention='ACN',
                                     normalization='N3D',
                                     condon_shortley=False)
    signal.channel_convention = 'FuMa'
    assert signal.channel_convention == 'FuMa'

    data_fuma = data[:, [0, 3, 1, 2], :]

    np.testing.assert_equal(data_fuma,
                            signal.time)


def test_spherical_harmonic_signal_renormalize():
    data = np.ones((1, 4, 2))
    signal = SphericalHarmonicSignal(data,
                                     44100, basis_type='real',
                                     channel_convention='ACN',
                                     normalization='N3D',
                                     condon_shortley=False)
    signal.normalization = 'maxN'
    data_ref = np.array([[np.sqrt(1 / 2), np.sqrt(1 / 2)],
                         [np.sqrt(1 / 3), np.sqrt(1 / 3)],
                         [np.sqrt(1 / 3), np.sqrt(1 / 3)],
                         [np.sqrt(1 / 3), np.sqrt(1 / 3)]]).reshape(1, 4, 2)

    np.testing.assert_equal(signal.time,
                            data_ref)



"""
Tests for modal strength function.
"""
from scipy.io import loadmat
import spharpy.spherical as sh
import numpy as np


def test_modal_strength_open():
    n_max = 5
    n_bins = 128
    k = np.linspace(0.5, 15, n_bins)
    fname = './tests/data/modal_strength_open.mat'
    reference = loadmat(fname)['matrix']
    bn = sh.modal_strength(n_max, k, arraytype='open')
    np.testing.assert_allclose(bn, reference)


def test_modal_strength_rigid():
    n_max = 5
    n_bins = 128
    k = np.linspace(0.5, 15, n_bins)
    # ref data has wrong sign (calculated as i^(n-1) instead of i^(n+1)

    fname = './tests/data/modal_strength_rigid.mat'
    reference = -loadmat(fname)['matrix']
    bn = sh.modal_strength(n_max, k, arraytype='rigid')
    np.testing.assert_allclose(bn, reference)



"""
This module offers plotting functions for spherical data based on Matplotlib.

Adjusting the Plot Layout
-------------------------

A common issue with matplotlib 3D axes objects is that layout-options like
:func:`matplotlib.pyplot.tight_layout` or
:ref:`constrained layout <matplotlib:constrainedlayout_guide>`
don't always work correctly.

This can cause elements to overlap, such as colorbar and axes.
To prevent this, spharpy plot functions enable passing a list of axes for
the plot itself and the colorbar. The best way to handle the layout of a
figure is :class:`matplotlib.gridspec.GridSpec`.


Example
~~~~~~~

.. plot::

    >>> import spharpy
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> from matplotlib.gridspec import GridSpec
    >>>
    >>> coords = spharpy.samplings.equal_area(n_max=0, n_points=500)
    >>> data = np.sin(coords.colatitude) * np.cos(coords.azimuth)
    >>>
    >>> # Create figure and GridSpec
    >>> fig = plt.figure(figsize=(9, 7))
    >>> gs = GridSpec(nrows=1, ncols=2, width_ratios=[20, 1], wspace=0.3)
    >>>
    >>> # Create subplot axes for plot and colorbar
    >>> ax = fig.add_subplot(gs[0], projection='3d')
    >>> cax = fig.add_subplot(gs[1])
    >>>
    >>> # Plot
    >>> spharpy.plot.balloon(coords, data, ax=[ax, cax])
    >>> plt.show()
"""

from .spatial import (
    scatter,
    pcolor_map,
    pcolor_sphere,
    balloon,
    balloon_wireframe,
    voronoi_cells_sphere,
    contour,
    contour_map,
    MidpointNormalize,
)

from .cmap import phase_twilight


__all__ = [
    'scatter',
    'pcolor_map',
    'pcolor_sphere',
    'balloon',
    'balloon_wireframe',
    'voronoi_cells_sphere',
    'contour',
    'contour_map',
    'MidpointNormalize',
    'phase_twilight',
]



"""Tests for sh channel indexing."""

import spharpy.spherical as sh
import numpy as np


def test_acn_to_nm_single_val():
    n, m = sh.acn_to_nm(0)
    assert n == 0
    assert m == 0

    n, m = sh.acn_to_nm(2)
    assert n == 1
    assert m == 0

    n, m = sh.acn_to_nm(1)
    assert n == 1
    assert m == -1


def test_nm_to_acn_single_val():
    acn = sh.nm_to_acn(0, 0)
    assert acn == 0

    acn = sh.nm_to_acn(1, 0)
    assert acn == 2

    acn = sh.nm_to_acn(1, -1)
    assert acn == 1


def test_acn_to_nm_array():
    n_ref = np.array([0, 1, 1, 1])
    m_ref = np.array([0, -1, 0, 1])

    n_max = 1
    n_sh = (n_max + 1)**2
    acn = np.arange(0, n_sh)
    n, m = sh.acn_to_nm(acn)

    np.testing.assert_equal(n, n_ref)
    np.testing.assert_equal(m, m_ref)


def test_nm_to_acn_array():
    n = np.array([0, 1, 1, 1])
    m = np.array([0, -1, 0, 1])

    n_max = 1
    n_sh = (n_max + 1)**2
    acn_ref = np.arange(0, n_sh)

    acn = sh.nm_to_acn(n, m)

    np.testing.assert_equal(acn, acn_ref)


def test_identity_matrix_n_nm():
    reference = np.array([[1, 0, 0, 0, 0, 0, 0, 0, 0],
                          [0, 1, 1, 1, 0, 0, 0, 0, 0],
                          [0, 0, 0, 0, 1, 1, 1, 1, 1]])

    identity = sh.sph_identity_matrix(2, matrix_type='n-nm')
    np.testing.assert_allclose(reference, identity)


def test_sid_indexing():
    n_max = 2
    reference_n = np.array([0, 1, 1, 1, 2, 2, 2, 2, 2])
    reference_m = np.array([0, 1, -1, 0, 2, -2, 1, -1, 0])

    sid_n, sid_m = sh.sid(n_max)

    np.testing.assert_equal(reference_n, sid_n)
    np.testing.assert_equal(reference_m, sid_m)


def test_sid_to_acn():
    n_max = 2
    # indexing starts at 0 here, reference was calculated
    # with indexing starting at 1.
    reference_acn = np.array([1, 3, 4, 2, 6, 8, 9, 7, 5]) - 1

    acn_indices = sh.sid_to_acn(n_max)
    np.testing.assert_equal(reference_acn, acn_indices)


def test_nm_to_fuma_single_val():
    fuma = sh.nm_to_fuma(0, 0)
    assert fuma == 0

    fuma = sh.nm_to_fuma(1, 0)
    assert fuma == 1

    fuma = sh.nm_to_fuma(1, -1)
    assert fuma == 3


def test_nm_to_fuma_array():
    n = np.array([0, 1, 1])
    m = np.array([0, 0, -1])

    fuma = sh.nm_to_fuma(n, m)
    np.testing.assert_equal(np.array([0, 1, 3], int), fuma)


def test_fuma_to_nm_single_val():
    n, m = sh.fuma_to_nm(0)
    assert n == 0
    assert m == 0

    n, m = sh.fuma_to_nm(1)
    assert n == 1
    assert m == 0

    n, m = sh.fuma_to_nm(3)
    assert n == 1
    assert m == -1


def test_fuma_to_nm_array():
    n_ref = np.array([0, 1, 1])
    m_ref = np.array([0, 0, -1])

    n, m = sh.fuma_to_nm(np.array([0, 1, 3]))
    np.testing.assert_equal(n_ref, n)
    np.testing.assert_equal(m_ref, m)



"""Beamforming methods for spherical harmonic signals."""

import itertools
import numpy as np
import numpy.polynomial as poly
from scipy.linalg import eigh
from scipy.special import factorial

import spharpy.special as special
from spharpy.spherical import sph_identity_matrix


def dolph_chebyshev_weights(
        n_max,
        design_parameter,
        design_criterion='sidelobe'):
    r"""
    Calculate the weights for a spherical Dolph-Chebyshev beamformer.

    The design criterion can either be a desired side-lobe attenuation or a
    desired main-lobe width. Once one criterion is chosen, the other will
    become a dependent property which will be chosen accordingly [#]_.

    Parameters
    ----------
    n_max : int
        Spherical harmonic order. Must be an integer greater than zero.
    design_parameter : float
        This can either be the desired side-lobe attenuation or the width of
        the main-lobe in radians.
    design_criterion : str
        Can be either ``'sidelobe'`` or ``'mainlobe'`` (see
        `design_parameter` above). The default is ``'sidelobe'``.

    Returns
    -------
    weights : array-like, float
        An array containing the :math:`(n_\mathrm{max} + 1)^2` beamformer
        weights.

    References
    ----------
    .. [#]  A. Koretz and B. Rafaely, “Dolph-Chebyshev beampattern design for
            spherical arrays,” IEEE Transactions on Signal Processing, vol. 57,
            no. 6, pp. 2417-2420, 2009.

    Examples
    --------
    Apply Dolph-Chebyshev beamformers with a side-lobe
    attenuation of 50 dB to a plane wave from zero degrees azimuth.

    .. plot::

        >>> import spharpy
        >>> import pyfar as pf
        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>>
        >>> # use real valued spherical harmonics of order 7
        >>> definition = spharpy.SphericalHarmonicDefinition(n_max=7)
        >>>
        >>> # define the direction of the incident sound field
        >>> # (0 degree azimuth)
        >>> soundfield = spharpy.SphericalHarmonics.from_definition(
        ...     definition, pf.Coordinates(1, 0, 0))
        >>> a_nm = soundfield.basis
        >>>
        >>> # beamforming
        >>> # a) generate 500 steering vectors
        >>> steering = spharpy.SphericalHarmonics.from_definition(
        ...     definition, pf.Coordinates.from_spherical_elevation(
        ...         np.linspace(0, 2*np.pi, 500), 0, 1))
        >>> Y_steering = steering.basis
        >>>
        >>> # b) design beamformer weights
        >>> R = 10**(50/20)
        >>> d_nm = spharpy.beamforming.dolph_chebyshev_weights(
        ...     definition.n_max, R, design_criterion='sidelobe')
        >>>
        >>> # c) appling beamformer weights yields 500 beamformers
        >>> beamformer = np.squeeze(Y_steering @ np.diag(d_nm))
        >>>
        >>> # d) apply beamformers to the soundfield
        >>> beamformer_output = beamformer @ a_nm.T
        >>>
        >>> # plot soundfield evaluated with beamformers
        >>> ax = plt.axes(projection='polar')
        >>> ax.plot(steering.coordinates.azimuth,
        ...         20*np.log10(np.abs(beamformer_output)))
        >>> ax.set_rticks([-50, -25, 0])
        >>> ax.set_theta_zero_location('N')
        >>> ax.set_xlabel('Azimuth in degree')
    """
    M = 2*n_max
    if design_criterion == 'sidelobe':
        R = design_parameter
        x0 = np.cosh((1/M) * np.arccosh(R))
    elif design_criterion == 'mainlobe':
        theta0 = design_parameter
        x0 = np.cos(np.pi/2/M) / np.cos(theta0/2)
        R = np.cosh(M * np.arccosh(x0))
    else:
        raise ValueError("This design criterion is not available.")

    t_2N = special.chebyshev_coefficients(2*n_max)

    P_N = np.zeros((n_max+1, n_max+1))
    for n in range(n_max+1):
        P_N[0:n+1, n] = special.legendre_coefficients(n)

    d_n = np.zeros(n_max+1)
    for n in range(n_max+1):
        temp = 0
        for i in range(n+1):
            for j in range(n_max+1):
                for m in range(j+1):
                    temp += (1-(-1)**(m+i+1))/(m+i+1) * \
                        factorial(j)/(factorial(m)*factorial(j-m)) * \
                        (1/2**j)*t_2N[2*j]*P_N[i, n]*x0**(2*j)
        d_n[n] = (2*np.pi/R)*temp

    return sph_identity_matrix(n_max, matrix_type='n-nm').T @ d_n


def rE_max_weights(n_max, normalize=True):
    r"""
    Calculate weights for a max :math:`\mathrm{r}_\mathrm{E}` beamformer.

    The weights that maximize the length of the energy vector.
    This is most often used in Ambisonics decoding [#]_.

    Parameters
    ----------
    n_max : int
        Spherical harmonic order. Must be an integer greater than zero.
    normalize : bool
        If ``True``, the weights will be normalized such that the complex
        amplitude of a plane wave is not distorted. The default is ``True``.

    Returns
    -------
    weights : array-like, float
        An array containing the :math:`(n_\mathrm{max} + 1)^2` beamformer
        weights.

    References
    ----------
    .. [#]  J. Daniel, J.-B. Rault, and J.-D. Polack, “Ambisonics Encoding of
            Other Audio Formats for Multiple Listening Conditions,” in 105th
            Convention of the Audio Engineering Society, 1998, vol. 3.

    Examples
    --------
    Apply max :math:`\mathrm{r}_\mathrm{E}` beamformers to a plane wave from
    zero degrees azimuth.

    .. plot::

        >>> import spharpy
        >>> import pyfar as pf
        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>>
        >>> # use real valued spherical harmonics of order 7
        >>> definition = spharpy.SphericalHarmonicDefinition(n_max=7)
        >>>
        >>> # define the direction of the incident sound field
        >>> # (0 degree azimuth)
        >>> soundfield = spharpy.SphericalHarmonics.from_definition(
        ...     definition, pf.Coordinates(1, 0, 0))
        >>> a_nm = soundfield.basis
        >>>
        >>> # beamforming
        >>> # a) generate 500 steering vectors
        >>> steering = spharpy.SphericalHarmonics.from_definition(
        ...     definition, pf.Coordinates.from_spherical_elevation(
        ...         np.linspace(0, 2*np.pi, 500), 0, 1))
        >>> Y_steering = steering.basis
        >>>
        >>> # b) design beamformer weights
        >>> d_nm = spharpy.beamforming.rE_max_weights(definition.n_max)
        >>>
        >>> # c) appling beamformer weights yields 500 beamformers
        >>> beamformer = np.squeeze(Y_steering @ np.diag(d_nm))
        >>>
        >>> # d) apply beamformers to the soundfield
        >>> beamformer_output = beamformer @ a_nm.T
        >>>
        >>> # plot soundfield evaluated with beamformers
        >>> ax = plt.axes(projection='polar')
        >>> ax.plot(steering.coordinates.azimuth,
        ...         20*np.log10(np.abs(beamformer_output)))
        >>> ax.set_rticks([-50, -25, 0])
        >>> ax.set_theta_zero_location('N')
        >>> ax.set_xlabel('Azimuth in degree')
    """
    leg = poly.legendre.Legendre.basis(n_max+1)
    P_n_root = poly.legendre.legroots(leg.coef)
    max_root = np.max(np.abs(P_n_root))
    g_n = np.zeros(n_max+1)
    for n in range(n_max+1):
        leg = poly.legendre.Legendre.basis(n)
        g_n[n] = leg(max_root)

    if normalize:
        g_n = normalize_beamforming_weights(g_n, n_max)

    return sph_identity_matrix(n_max).T @ g_n


def maximum_front_back_ratio_weights(n_max, normalize=True):
    r"""
    Compute weights that maximize the front-back ratio of the beam pattern.

    This is also often referred to as the super-cardioid beam pattern.
    The weights are calculated from an eigenvalue problem [#]_. For high
    spherical harmonic orders, the eigenvalue problem may not be feasible and
    a solution will not be found.

    Parameters
    ----------
    n_max : int
        The spherical harmonic order. Must be an integer greater than zero.
    normalize : bool
        If ``True``, the weights will be normalized such that the complex
        amplitude of a plane wave is not distorted. The default is ``True``.

    Returns
    -------
    weights : array-like, float
        An array containing the :math:`(n_\mathrm{max} + 1)^2` beamformer
        weights.

    References
    ----------
    .. [#]  B. Rafaely, Fundamentals of Spherical Array Processing,
            Springer, 2015.

    Examples
    --------
    Apply beamformers maximizing the front-back ratio to a plane wave from
    zero degrees azimuth.

    .. plot::

        >>> import spharpy
        >>> import pyfar as pf
        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>>
        >>> # use real valued spherical harmonics of order 5
        >>> definition = spharpy.SphericalHarmonicDefinition(n_max=5)
        >>>
        >>> # define the direction of the incident sound field
        >>> # (0 degree azimuth)
        >>> soundfield = spharpy.SphericalHarmonics.from_definition(
        ...     definition, pf.Coordinates(1, 0, 0))
        >>> a_nm = soundfield.basis
        >>>
        >>> # beamforming
        >>> # a) generate 500 steering vectors
        >>> steering = spharpy.SphericalHarmonics.from_definition(
        ...     definition, pf.Coordinates.from_spherical_elevation(
        ...         np.linspace(0, 2*np.pi, 500), 0, 1))
        >>> Y_steering = steering.basis
        >>>
        >>> # b) design beamformer weights
        >>> R = 10**(50/20)
        >>> d_nm = spharpy.beamforming.maximum_front_back_ratio_weights(
        ...     definition.n_max)
        >>>
        >>> # c) appling beamformer weights yields 500 beamformers
        >>> beamformer = np.squeeze(Y_steering @ np.diag(d_nm))
        >>>
        >>> # d) apply beamformers to the soundfield
        >>> beamformer_output = beamformer @ a_nm.T
        >>>
        >>> # plot soundfield evaluated with beamformers
        >>> ax = plt.axes(projection='polar')
        >>> ax.plot(steering.coordinates.azimuth,
        ...         20*np.log10(np.abs(beamformer_output)))
        >>> ax.set_rticks([-75, -50, -25, 0])
        >>> ax.set_theta_zero_location('N')
        >>> ax.set_xlabel('Azimuth in degree')
    """
    P_N = np.zeros((n_max+1, n_max+1))
    for n in range(n_max+1):
        P_N[0:n+1, n] = special.legendre_coefficients(n)

    Ann = np.zeros((n_max+1, n_max+1))
    Bnn = np.zeros((n_max+1, n_max+1))
    for n in range(n_max+1):
        for n_dash in range(n_max+1):
            const = 1/8/np.pi * (2*n+1) * (2*n_dash+1)
            temp = sum(
                1 / (q+ll+1) * P_N[q, n] * P_N[ll, n_dash]
                for q, ll in itertools.product(range(n+1), range(n_dash+1)))
            Ann[n, n_dash] = temp * const

            temp = sum(
                ((-1) ** (q+ll)) / (q+ll+1) * P_N[q, n] * P_N[ll, n_dash]
                for q, ll in itertools.product(range(n+1), range(n_dash+1)))
            Bnn[n, n_dash] = temp * const

    try:
        eigenvals, eigenvectors = eigh(Ann, Bnn)
    except np.linalg.LinAlgError as e:
        raise RuntimeError(
            'Eigenvalue decomposition did not converge. '
            'Try reducing the spherical harmonic order.') from e
    f_n = eigenvectors[:, np.argmax(np.real(eigenvals))]
    if normalize:
        f_n = normalize_beamforming_weights(f_n, n_max)
    else:
        f_n /= np.sign(f_n[0])

    return sph_identity_matrix(n_max).T @ f_n


def normalize_beamforming_weights(weights, n_max):
    r"""
    Normalize the beamforming weights.

    The weights are normalized such that the complex amplitude of a
    plane wave is not distorted.

    Parameters
    ----------
    weights : array-like, float
        An array containing the beamforming weights. The array must be of
        shape :math:`(\dots, (n_\mathrm{max}+1)^2)`.
    n_max : int
        The spherical harmonic order. Must be an integer greater than zero.

    Returns
    -------
    weights : array-like, float
        An array containing the normalized beamforming weights. The shape
        of the output weights is the same as the shape of the input weights.

    Examples
    --------
    Calculate and normalize hann window function based beamforming weights.

    >>> import spharpy
    >>> from scipy.signal.windows import hann
    ...
    >>> tapering_window = hann(2*(N+1)+1)[N+1:-1], N)
    >>> h_n = spharpy.beamforming.normalize_beamforming_weights(
    ...     tapering_window, N)

    """
    return weights / np.dot(weights, 2*np.arange(0, n_max+1)+1) * (4*np.pi)
